'''
Copyright (C) 2025  ElaadNL

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

-------------------------------------------------------------------------------
'''
from .modbus import ModbusDevice


class Battery:
    '''Base battery class

    Arguments:
        capacity (int): the capacity of the battery in Wh
        max_charge_power (int): the maximum charge power of the battery in W
        max_discharge_power (int): the maximum discharge power of the battery
                                   in W
        soc (int): the state of charge of the battery in %
    '''

    def __init__(
            self,
            capacity: int = None,
            max_charge_power: int = None,
            max_discharge_power: int = None,
            soc: int = None
    ):
        self._capacity = capacity
        self._max_charge_power = max_charge_power
        self._max_discharge_power = max_discharge_power
        self._soc = soc
        self._client = None

    def set_power_setpoint(self, setpoint: int):
        raise NotImplementedError("Not implemented")

    def set_current_setpoint(self, setpoint: int):
        raise NotImplementedError("Not implemented")

    def set_power_setpoint_phase(self, setpoint: int, phase: int):
        raise NotImplementedError("Not implemented")

    def set_current_setpoint_phase(self, setpoint: int, phase: int):
        raise NotImplementedError("Not implemented")

    def get_active_power_total(self):
        raise NotImplementedError("Not implemented")

    def get_active_current_total(self):
        raise NotImplementedError("Not implemented")

    def get_active_power_phase(self, phase: int):
        raise NotImplementedError("Not implemented")

    def get_active_current_phase(self, phase: int):
        raise NotImplementedError("Not implemented")

    def get_phase_voltage(self, phase: int):
        raise NotImplementedError("Not implemented")

    def get_battery_soc(self):
        raise NotImplementedError("Not implemented")


class ModbusBattery(Battery, ModbusDevice):
    '''Modbus battery class

    Arguments:
        modbus_ip (str): the ip address of the modbus battery
        modbus_port (int): the port of the modbus battery
        timeout (int): the timeout of the modbus battery in seconds
        capacity (int): the capacity of the battery in Wh
        max_charge_power (int): the maximum charge power of the battery in W
        max_discharge_power (int): the maximum discharge power of the battery
                                   in W
        soc (int): the state of charge of the battery in %
    '''

    def __init__(
            self,
            modbus_ip: str,
            modbus_port: int = 502,
            timeout: int = 3,
            capacity: int = None,
            max_charge_power: int = None,
            max_discharge_power: int = None,
            soc: int = None
    ):
        Battery.__init__(
            self,
            capacity,
            max_charge_power,
            max_discharge_power,
            soc
        )
        ModbusDevice.__init__(
            self,
            modbus_ip,
            modbus_port,
            timeout
        )



"""
Copyright (C) 2025  ElaadNL

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

-------------------------------------------------------------------------------

This module defines the Sessy class, which represents a battery system with
various functionalities such as retrieving power status, phase voltage, state
of charge (SOC), power, energy status, and active strategy. It also allows
setting the active strategy and sending power or current setpoints to the
battery system. The Sessy class inherits from the Battery class and uses
aiohttp for asynchronous HTTP requests to interact with the battery system's
API.
"""

import warnings
from ..battery import Battery
from aiohttp import BasicAuth, ClientSession, ClientConnectionError, \
    ClientResponseError, ContentTypeError
from aiohttp.client_exceptions import InvalidURL
from aiohttp import ClientTimeout
from enum import Enum


class PowerStrategy(Enum):
    NOM = "POWER_STRATEGY_NOM"
    ROI = "POWER_STRATEGY_ROI"
    API = "POWER_STRATEGY_API"
    IDLE = "POWER_STRATEGY_IDLE"
    SESSY_CONNECT = "POWER_STRATEGY_SESSY_CONNECT"
    ECO = "POWER_STRATEGY_ECO"


class SessyInverter(Battery):
    '''Sessy inverter class

    Arguments:
        ip (str): the ip address of the Sessy battery
        username (str): the username for the Sessy battery
        password (str): the password for the Sessy battery
        capacity (int): the capacity of the battery in Wh
        max_charge_power (int): the maximum charge power of the battery in W
        max_discharge_power (int): the maximum discharge power of the battery
                                   in W
        soc (int): the state of charge of the battery in %
    '''

    def __init__(
            self,
            ip: str,
            username: str,
            password: str,
            capacity: int = None,
            max_charge_power: int = None,
            max_discharge_power: int = None,
            soc: int = None
    ):
        super().__init__(capacity, max_charge_power, max_discharge_power, soc)
        self._ip = ip
        self._soc = soc
        self._power = None
        self._voltage = {"phase1": None, "phase2": None,
                         "phase3": None, "total": None}
        self._current = {"phase1": None, "phase2": None,
                         "phase3": None, "total": None}
        self._url = f"http://{self._ip}/api/v1"
        self._username = username
        self._password = password
        self._client = ClientSession(
            auth=BasicAuth(self._username, self._password),
            timeout=ClientTimeout(total=60),
            raise_for_status=True
        )

    async def _get_power_status(self):
        ''' Get the power status of the inverter'''
        endpoint = f"{self._url}/power/status"
        try:
            async with self._client.get(endpoint) as response:
                data = await response.json()
                sessy_data = data.get('sessy')
                if sessy_data:
                    self._soc = sessy_data.get('state_of_charge') * 100
                    self._power = sessy_data.get('power')
                return data
        except (ClientConnectionError, ClientResponseError,
                ContentTypeError, InvalidURL) as e:
            print(f"Error retrieving power status: {e}")
            raise

    async def _get_phase_data(self, phase: int):
        ''' Get the phase data of the inverter'''
        endpoint = f"{self._url}/power/status"
        try:
            async with self._client.get(endpoint) as response:
                data = await response.json()
                phase_data = data.get(f"renewable_energy_phase{phase}")
                if phase_data:
                    phase_voltage = phase_data.get('voltage_rms')
                    phase_current = phase_data.get('current_rms')
                    self._voltage[f"phase{phase}"] = phase_voltage / 1000
                    self._current[f"phase{phase}"] = phase_current
        except (ClientConnectionError, ClientResponseError,
                ContentTypeError, InvalidURL) as e:
            print(f"Error retrieving power status: {e}")
            raise

    async def get_phase_voltage(self, phase: int):
        ''' Get the voltage of a specific phase of the inverter'''
        if phase not in [1, 2, 3]:
            raise ValueError("Invalid phase input, must be 1, 2 or 3")

        await self._get_phase_data(phase)
        return self._voltage[f"phase{phase}"]

    async def get_phase_current(self, phase: int):
        ''' Get the active current of a specific phase of the inverter'''
        if phase not in [1, 2, 3]:
            raise ValueError("Invalid phase input, must be 1, 2 or 3")

        await self._get_phase_data(phase)
        return self._current[f"phase{phase}"]

    async def get_battery_soc(self):
        ''' Get the state of charge of the battery'''
        await self._get_power_status()
        return self._soc

    async def get_active_power_total(self):
        ''' Get the total active power of the inverter'''
        await self._get_power_status()
        return self._power

    async def get_active_current_total(self):
        ''' Get the total active current of the inverter'''
        await self._get_power_status()
        self._current["total"] = \
            self._power / await self.get_phase_voltage(phase=1)
        return self._current["total"]

    async def get_active_strategy(self):
        ''' Get the active strategy of the inverter'''
        endpoint = f"{self._url}/power/active_strategy"
        try:
            async with self._client.get(endpoint) as response:
                return await response.json()
        except (ClientConnectionError, ClientResponseError,
                ContentTypeError, InvalidURL) as e:
            print(f"Error retrieving active strategy: {e}")
            raise

    async def set_active_strategy(self, strategy: PowerStrategy):
        ''' Set the active strategy of the battery system'''
        endpoint = f"{self._url}/power/active_strategy"
        payload = {"strategy": strategy.value}
        try:
            async with self._client.post(endpoint, json=payload) as response:
                if response.status == 200:
                    return True
        except (ClientConnectionError, ClientResponseError,
                ContentTypeError, InvalidURL) as e:
            print(f"Error setting active strategy: {e}")
            raise

    async def set_power_setpoint(self, setpoint: int):
        ''' Set the power setpoint of the battery system'''
        if (self._max_charge_power is None or
                self._max_discharge_power is None):
            warnings.warn("Max charge power or max discharge power not set, \
                          setpoint may exceed limits")
        elif setpoint > self._max_charge_power:
            warnings.warn(f"Power setpoint {setpoint} exceeds the maximum \
                          charge power {self._max_charge_power} \n\
                          Reverting to maximum charge power")
            setpoint = self._max_charge_power
        elif setpoint < -self._max_discharge_power:
            warnings.warn(f"Power setpoint {setpoint} exceeds the maximum \
                          discharge power {-self._max_discharge_power} \n\
                          Reverting to maximum discharge power")
            setpoint = -self._max_discharge_power

        endpoint = f"{self._url}/power/setpoint"
        payload = {"setpoint": setpoint}
        try:
            async with self._client.post(endpoint, json=payload) as response:
                if response.status == 200:
                    return True
        except (ClientConnectionError, ClientResponseError,
                ContentTypeError, InvalidURL) as e:
            print(f"Error sending setpoint: {e}")
            raise

    async def set_current_setpoint(self, setpoint: int):
        ''' Set the current setpoint of the battery system'''
        phase_voltage = await self.get_phase_voltage(phase=1)
        power_setpoint = setpoint * phase_voltage
        return await self.set_power_setpoint(power_setpoint)



from .batteries.deye_modbus import DeyeModbusInverter
from .batteries.deye_solarman import DeyeSolarmanInverter
from .batteries.sessy import SessyInverter
from .batteries.sun2000 import Sun2000Inverter
from .batteries.victron import VictronInverter
from .batteries.sma import SmaInverter
from .battery import Battery, ModbusBattery



"""
Copyright (C) 2025  ElaadNL

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

-------------------------------------------------------------------------------

This module provides a class to interact with the Sun2000 inverter using Modbus
protocol. The Sun2000Inverter class allows for reading and writing various
parameters and settings of the inverter, such as power setpoints, battery state
of charge, and phase voltages.

The class inherits from ModbusBattery and utilizes the Register, AccessType,
and DataType classes from the modbus module to define and access the Modbus
registers of the inverter.
"""

import warnings
from ..battery import ModbusBattery
from ..modbus import Register, AccessType, DataType
from enum import Enum


class Sun2000Inverter(ModbusBattery):
    '''Sun2000 inverter class

    Arguments:
        modbus_ip (str): the ip address of the modbus inverter
        modbus_port (int): the port of the modbus inverter
        timeout (int): the timeout of the modbus client in seconds
        capacity (int): the capacity of the inverter
        max_charge_power (int): the maximum charge power of the battery in W
        max_discharge_power (int): the maximum discharge power of the battery
                                   in W
        soc (int): the state of charge of the battery in %
    '''

    def __init__(
            self,
            modbus_ip: str,
            modbus_port: int = 502,
            timeout: int = 3,
            capacity: int = None,
            max_charge_power: int = None,
            max_discharge_power: int = None,
            soc: int = None
    ):
        super().__init__(
            modbus_ip, modbus_port, timeout, capacity, max_charge_power,
            max_discharge_power, soc)

    def init_power_setpoint_registers(self):
        '''Prepare the inverter to accept power setpoints'''
        self.write_register(
            self.Sun2000Registers.ChargeFromGridFunction.value, 1)
        self.write_register(
            self.Sun2000Registers.ForcibleChargeDischargeSettingMode.value, 1)

    def _set_charge_setpoint(self, setpoint: int):
        '''Set the charge setpoint of the inverter

        Arguments:
            setpoint (int): the charge setpoint to set
        '''
        self.write_register(
            self.Sun2000Registers.TargetSOC.value, 1000)
        self.write_register(
            self.Sun2000Registers.ForcibleChargePower.value, setpoint)
        self.write_register(
            self.Sun2000Registers.ForcibleChargeDischarge.value, 1)

    def _set_discharge_setpoint(self, setpoint: int):
        '''Set the discharge setpoint of the inverter

        Arguments:
            setpoint (int): the discharge setpoint to set
        '''
        self.write_register(
            self.Sun2000Registers.TargetSOC.value, 50)
        self.write_register(
            self.Sun2000Registers.ForcibleDischargePower.value, setpoint)
        self.write_register(
            self.Sun2000Registers.ForcibleChargeDischarge.value, 2)

    def set_power_setpoint(self, setpoint: int):
        '''Set the power setpoint of the inverter

        Arguments:
            setpoint (int): the power setpoint to set in W
        '''
        if (self._max_charge_power is None or
                self._max_discharge_power is None):
            warnings.warn("Max charge power or max discharge power not set, \
                          setpoint may exceed limits")
        elif setpoint > self._max_charge_power:
            warnings.warn(f"Power setpoint {setpoint} exceeds the maximum \
                          charge power {self._max_charge_power} \n\
                          Reverting to maximum charge power")
            setpoint = self._max_charge_power
        elif setpoint < -self._max_discharge_power:
            warnings.warn(f"Power setpoint {setpoint} exceeds the maximum \
                          discharge power {-self._max_discharge_power} \n\
                          Reverting to maximum discharge power")
            setpoint = -self._max_discharge_power

        if setpoint > 0:
            self._set_charge_setpoint(setpoint)
        elif setpoint < 0:
            self._set_discharge_setpoint(abs(setpoint))
        else:
            self.write_register(
                self.Sun2000Registers.ForcibleChargeDischarge.value, 0)

    def set_current_setpoint(self, setpoint: int):
        '''Set the current setpoint of the inverter

        Arguments:
            setpoint (int): the current setpoint to set in A
        '''
        phase_voltage = self.get_phase_voltage(1)
        power_setpoint = setpoint * phase_voltage
        self.set_power_setpoint(power_setpoint)

    def get_active_power_total(self):
        '''Get the total active power of the inverter

        Returns:
            float: the active power of the inverter
        '''
        return self.read_holding_register(
            self.Sun2000Registers.BatteryActivePower.value)

    def get_active_current_total(self):
        '''Get the total active current of the inverter

        Returns:
            float: the active current of the inverter
        '''
        return self.get_active_power_total() / self.get_phase_voltage(1)

    def get_battery_soc(self):
        '''Get the battery state of charge

        Returns:
            int: the battery state of charge
        '''
        return self.read_holding_register(
            self.Sun2000Registers.BatterySOC.value)

    def get_phase_voltage(self, phase: int):
        '''Get the phase voltage

        Arguments:
            phase (int): the phase to get the voltage from

        Returns:
            float: the phase voltage
        '''
        register = None
        if phase == 1:
            register = self.Sun2000Registers.PowerGridPhaseAVoltage.value
        elif phase == 2:
            register = self.Sun2000Registers.PowerGridPhaseBVoltage.value
        elif phase == 3:
            register = self.Sun2000Registers.PowerGridPhaseCVoltage.value
        if register:
            return self.read_holding_register(register)
        else:
            raise ValueError("Invalid phase")

    def get_phase_current(self, phase: int):
        '''Get the phase current

        Arguments:
            phase (int): the phase to get the current from

        Returns:
            float: the phase current
        '''
        register = None
        if phase == 1:
            register = self.Sun2000Registers.PowerGridPhaseACurrent.value
        elif phase == 2:
            register = self.Sun2000Registers.PowerGridPhaseBCurrent.value
        elif phase == 3:
            register = self.Sun2000Registers.PowerGridPhaseCCurrent.value
        if register:
            return self.read_holding_register(register)
        else:
            raise ValueError("Invalid phase")

    class Sun2000Registers(Enum):
        Model = Register(
            30000, "model", DataType.STRING, count=15,
            access_type=AccessType.READ)
        SerialNumber = Register(
            30015, "serial_number", DataType.STRING, count=10,
            access_type=AccessType.READ)
        ProductCode = Register(
            30025, "product_code", DataType.STRING, count=10,
            access_type=AccessType.READ)
        FirmwareVersion = Register(
            30035, "firmware_version", DataType.STRING, count=15,
            access_type=AccessType.READ)
        SoftwareVersion = Register(
            30050, "software_version", DataType.STRING, count=15,
            access_type=AccessType.READ)
        ProtocolVersion = Register(
            30068, "protocol_version", DataType.UINT32,
            access_type=AccessType.READ)
        ModelID = Register(
            30070, "model_id", DataType.UINT16, access_type=AccessType.READ)
        NumberOfStrings = Register(
            30071, "number_of_strings", DataType.UINT16,
            access_type=AccessType.READ)
        NumberOfMPPTs = Register(
            30072, "number_of_mppts", DataType.UINT16,
            access_type=AccessType.READ)
        RatedPower = Register(
            30073, "rated_power", DataType.UINT32, gain=1000,
            access_type=AccessType.READ)
        MaximumActivePowerR = Register(
            30075, "maximum_active_power", DataType.UINT32, gain=1000,
            access_type=AccessType.READ)
        MaximumApparantPower = Register(
            30077, "maximum_aparrant_power", DataType.UINT32, gain=1000,
            access_type=AccessType.READ)
        RealTimeMaximumReactivePowerFeedToGrid = Register(
            30079, "real_time_maximum_reactive_power_feed_to_grid",
            DataType.INT32, gain=1000, access_type=AccessType.READ)
        RealTimeMaximumReactivePowerAbsorbedFromGrid = Register(
            30081, "real_time_maximum_reactive_power_absorbed_from_grid",
            DataType.INT32, gain=1000, access_type=AccessType.READ)
        MaximumActiveCapability = Register(
            30083, "maximum_active_capability", DataType.UINT32, gain=1000,
            access_type=AccessType.READ)
        MaximumApparentCapability = Register(
            30085, "maximum_apparent_capability", DataType.UINT32, gain=1000,
            access_type=AccessType.READ)
        PowerGridPhaseAVoltage = Register(
            32069, "power_grid_phase_a_voltage", DataType.UINT16, gain=10,
            access_type=AccessType.READ)
        PowerGridPhaseBVoltage = Register(
            32070, "power_grid_phase_b_voltage", DataType.UINT16, gain=10,
            access_type=AccessType.READ)
        PowerGridPhaseCVoltage = Register(
            32071, "power_grid_phase_c_voltage", DataType.UINT16, gain=10,
            access_type=AccessType.READ)
        PowerGridPhaseACurrent = Register(
            32072, "power_grid_phase_a_current", DataType.INT32, gain=1000,
            count=2, access_type=AccessType.READ)
        PowerGridPhaseBCurrent = Register(
            32074, "power_grid_phase_b_current", DataType.INT32, gain=1000,
            count=2, access_type=AccessType.READ)
        PowerGridPhaseCCurrent = Register(
            32076, "power_grid_phase_c_current", DataType.INT32, gain=1000,
            count=2, access_type=AccessType.READ)
        ActivePower = Register(
            32080, "active_power", DataType.INT32, gain=1000,
            access_type=AccessType.READ)
        ReactivePower = Register(
            32082, "reactive_power", DataType.INT32, gain=1000,
            access_type=AccessType.READ)
        PowerFactorR = Register(
            32084, "power_factor", DataType.INT16, gain=1000,
            access_type=AccessType.READ)
        GridFrequency = Register(
            32085, "grid_frequency", DataType.UINT16, gain=100,
            access_type=AccessType.READ)
        InverterEfficiency = Register(
            32086, "inverter_efficiency", DataType.UINT16, gain=100,
            access_type=AccessType.READ)
        InternalTemperature = Register(
            32087, "internal_temperature", DataType.INT16, gain=10,
            access_type=AccessType.READ)
        InsulationImpedanceValue = Register(
            32088, "insulation_impedance_value", DataType.UINT16, gain=1000,
            access_type=AccessType.READ)
        DeviceStatus = Register(
            32089, "device_status", DataType.UINT16,
            access_type=AccessType.READ)
        FaultCode = Register(
            32090, "fault_code", DataType.UINT16, access_type=AccessType.READ)
        StartupTime = Register(
            32091, "startup_time", DataType.UINT32,
            access_type=AccessType.READ)
        ShutdownTime = Register(
            32093, "shutdown_time", DataType.UINT32,
            access_type=AccessType.READ)
        ActivePowerFast = Register(
            32095, "active_power_fast", DataType.INT32, gain=1000,
            access_type=AccessType.READ)
        SystemTime = Register(
            40000, "system_time", DataType.UINT32, count=2,
            access_type=AccessType.READ_WRITE)
        Q_UCharacteristicCurveModel = Register(
            40037, "q_u_characteristic_curve_model", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        Q_USchedulingTriggerPowerPercentage = Register(
            40038, "q_u_scheduling_trigger_power_percentage", DataType.INT16,
            access_type=AccessType.READ_WRITE)
        ActivePowerFixedValueDerating = Register(
            40120, "active_power_fixed_value_derating", DataType.UINT16,
            gain=10, access_type=AccessType.READ_WRITE)
        PowerFactorW = Register(
            40122, "power_factor", DataType.INT16, gain=1000,
            access_type=AccessType.READ_WRITE)
        ReactivePowerCompensation = Register(
            40123, "reactive_power_compensation", DataType.INT16, gain=1000,
            access_type=AccessType.READ_WRITE)
        ReactivePowerAdjustmentTime = Register(
            40124, "reactive_power_adjustment_time", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        ActivePowerPercentageDerating = Register(
            40125, "active_power_percentage_derating", DataType.INT16,
            gain=10, access_type=AccessType.READ_WRITE)
        ActivePowerFixedValueDerating2 = Register(
            40126, "active_power_fixed_value_derating", DataType.UINT32,
            count=2, access_type=AccessType.READ_WRITE)
        ReactivePowerCompensationAtNight = Register(
            40128, "reactive_power_compensation_at_night", DataType.INT16,
            gain=1000, access_type=AccessType.READ_WRITE)
        FixedReactivePowerAtNight = Register(
            40129, "fixed_reactive_power_at_night", DataType.INT32, gain=1000,
            count=2, access_type=AccessType.READ_WRITE)
        CharacteristiccurveReactivepoweradjustmentTime = Register(
            40196, "characteristiccurve_reactivepoweradjustment_time",
            DataType.UINT16, access_type=AccessType.READ_WRITE)
        PercentApparentPower = Register(
            40197, "percent_apparent_power", DataType.UINT16, gain=10,
            access_type=AccessType.READ_WRITE)
        Q_USchedulingExitPowerPercentage = Register(
            40198, "q_u_scheduling_exit_power_percentage", DataType.INT16,
            access_type=AccessType.READ_WRITE)
        ActivePowerPercentageControl = Register(
            40199, "active_power_percentage_control", DataType.INT16, gain=10,
            access_type=AccessType.READ_WRITE)
        PowerOn = Register(
            40200, "power_on", DataType.UINT16, access_type=AccessType.WRITE)
        Shutdown = Register(
            40201, "shutdown", DataType.UINT16, access_type=AccessType.WRITE)
        Reset = Register(
            40205, "reset", DataType.UINT16, access_type=AccessType.WRITE)
        GridStandardCode = Register(
            42000, "grid_standard_code", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        OutputMode = Register(
            42001, "output_mode", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        VoltageLevel = Register(
            42002, "voltage_level", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        FrequencyLevel = Register(
            42003, "frequency_level", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        RemotePowerScheduling = Register(
            42014, "remote_power_scheduling", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        ReactivePowerVariationGradient = Register(
            42015, "reactive_power_variation_gradient", DataType.UINT32,
            gain=1000, count=2, access_type=AccessType.READ_WRITE)
        ActivePowerGradient = Register(
            42017, "active_power_gradient", DataType.UINT32, gain=1000,
            count=2, access_type=AccessType.READ_WRITE)
        SchedulingInstructionMaintenanceTime = Register(
            42019, "scheduling_instruction_maintenance_time", DataType.UINT32,
            count=2, access_type=AccessType.READ_WRITE)
        MaximumApparentPower = Register(
            42021, "maximum_apparent_power", DataType.UINT32, gain=1000,
            count=2, access_type=AccessType.READ_WRITE)
        MaximumActivePowerW = Register(
            42023, "maximum_active_power", DataType.UINT32, gain=1000, count=2,
            access_type=AccessType.READ_WRITE)
        ApparentPowerReference = Register(
            42025, "apparent_power_reference", DataType.UINT32, gain=1000,
            count=2, access_type=AccessType.READ_WRITE)
        ActivePowerReference = Register(
            42027, "active_power_reference", DataType.UINT32, gain=1000,
            count=2, access_type=AccessType.READ_WRITE)
        ActivePowerGradientOfPowerStation = Register(
            42029, "active_power_gradient_of_power_station", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        FilteringTimeOfTheAverageActivePowerOfThePowerStation = Register(
            42030,
            "filtering_time_of_the_average_active_power_of_the_power_station",
            DataType.UINT32, count=2, access_type=AccessType.READ_WRITE)
        PF_UVoltageDetectionFilterTime = Register(
            42032, "pf_u_voltage_detection_filter_time", DataType.UINT16,
            gain=10, access_type=AccessType.READ_WRITE)
        FrequencyDetectionFilterTime = Register(
            42037, "frequency_detection_filter_time", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        FrequencyActiveDeratingRecoveryDelayTime = Register(
            42040, "frequency_active_derating_recovery_delay_time",
            DataType.UINT16, access_type=AccessType.READ_WRITE)
        EffectiveDelayTimeOfActiveFrequencyDerating = Register(
            42041, "effective_delay_time_of_active_frequency_derating",
            DataType.UINT16, access_type=AccessType.READ_WRITE)
        FrequencyActiveDeratingHysteresisLoop = Register(
            42042, "frequency_active_derating_hysteresis_loop",
            DataType.UINT16, access_type=AccessType.READ_WRITE)
        FMControlResponseDeadZone = Register(
            42043, "fm_control_response_dead_zone", DataType.UINT16, gain=1000,
            access_type=AccessType.READ_WRITE)
        PQMode = Register(
            42046, "pq_mode", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        PanelType = Register(
            42047, "panel_type", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        PIDCompensationDirection = Register(
            42048, "pid_compensation_direction", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        StringConnectionMode = Register(
            42049, "string_connection_mode", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        IsolationSettings = Register(
            42050, "isolation_settings", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        FrequencyModulationControlPowerVariationGradient = Register(
            42051, "frequency_modulation_control_power_variation_gradient",
            DataType.UINT16, access_type=AccessType.READ_WRITE)
        FMControlPowerVariationLimit = Register(
            42052, "fm_control_power_variation_limit", DataType.UINT16,
            gain=10, access_type=AccessType.READ_WRITE)
        FMControlDelayResponseTime = Register(
            42053, "fm_control_delay_response_time", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        MPPTMultimodalScanning = Register(
            42054, "mppt_multimodal_scanning", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        MPPTScanningInterval = Register(
            42055, "mppt_scanning_interval", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        MPPTPredictedPower = Register(
            42056, "mppt_predicted_power", DataType.UINT32, gain=1000, count=2,
            access_type=AccessType.READ)
        AutomaticPowerGridFaultRecovery = Register(
            42061, "automatic_power_grid_fault_recovery", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        PowerLimit0Shutdown = Register(
            42062, "power_limit_0_shutdown", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        AutomaticShutoffOfCommunicationLinkDisconnection = Register(
            42063, "automatic_shutoff_of_communication_link_disconnection",
            DataType.UINT16, access_type=AccessType.READ_WRITE)
        CommunicationResumesAutomaticPowerOn = Register(
            42064, "communication_resumes_automatic_power_on", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        PowerQualityOptimizationMode = Register(
            42065, "power_quality_optimization_mode", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        RCDEnhancement = Register(
            42066, "rcd_enhancement", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        NoTimeWork = Register(
            42067, "no_time_work", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        NightPIDProtection = Register(
            42069, "night_pid_protection", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        ReactivePowerParameterTakesEffectAtNight = Register(
            42070, "reactive_power_parameter_takes_effect_at_night",
            DataType.UINT16, access_type=AccessType.READ_WRITE)
        CommunicationDisconnectionDetectionTime = Register(
            42072, "communication_disconnection_detection_time",
            DataType.UINT16, access_type=AccessType.READ_WRITE)
        AFCI = Register(
            42073, "afci", DataType.UINT16, access_type=AccessType.READ_WRITE)
        AFCIDetectionAdaptationMode = Register(
            42074, "afci_detection_adaptation_mode", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        CommunicationLinkDisconnectionFailureProtection = Register(
            42075, "communication_link_disconnection_failure_protection",
            DataType.UINT16, access_type=AccessType.READ_WRITE)
        FailSafeActivePowerMode = Register(
            42076, "fail_safe_active_power_mode", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        ActivePowerLimitForFailProtection = Register(
            42077, "active_power_limit_for_fail_protection", DataType.UINT32,
            gain=10, count=2, access_type=AccessType.READ_WRITE)
        FailSafeReactivePowerMode = Register(
            42079, "fail_safe_reactive_power_mode", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        FrequencyChangeRateProtection = Register(
            42080, "frequency_change_rate_protection", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        FrequencyChangeRateProtectionPoint = Register(
            42081, "frequency_change_rate_protection_point", DataType.UINT16,
            gain=10, access_type=AccessType.READ_WRITE)
        FrequencyChangeRateProtectionTime = Register(
            42082, "frequency_change_rate_protection_time", DataType.UINT16,
            gain=10, access_type=AccessType.READ_WRITE)
        FailProtectionReactivePowerLimit = Register(
            42083, "fail_protection_reactive_power_limit", DataType.INT16,
            gain=1000, access_type=AccessType.READ_WRITE)
        PowerOnVoltageUpperLimit = Register(
            42084, "power_on_voltage_upper_limit", DataType.UINT16, gain=10,
            access_type=AccessType.READ_WRITE)
        LowerLimitOfGridConnectedPowerOnVoltage = Register(
            42085, "lower_limit_of_grid_connected_power_on_voltage",
            DataType.UINT16, gain=10, access_type=AccessType.READ_WRITE)
        UpperLimitOfGridComingStartupFrequency = Register(
            42086, "upper_limit_of_grid_coming_startup_frequency",
            DataType.UINT16, gain=100, access_type=AccessType.READ_WRITE)
        LowerLimitOfPowerOnFrequency = Register(
            42087, "lower_limit_of_power_on_frequency", DataType.UINT16,
            gain=100, access_type=AccessType.READ_WRITE)
        UpperLimitOfPowerGridReconnectionVoltage = Register(
            42088, "upper_limit_of_power_grid_reconnection_voltage",
            DataType.UINT16, gain=10, access_type=AccessType.READ_WRITE)
        LowerLimitOfPowerGridReconnectionVoltage = Register(
            42089, "lower_limit_of_power_grid_reconnection_voltage",
            DataType.UINT16, gain=10, access_type=AccessType.READ_WRITE)
        UpperLimitOfPowerGridReconnectionFrequency = Register(
            42090, "upper_limit_of_power_grid_reconnection_frequency",
            DataType.UINT16, gain=100, access_type=AccessType.READ_WRITE)
        LowerLimitOfPowerGridReconnectionFrequency = Register(
            42091, "lower_limit_of_power_grid_reconnection_frequency",
            DataType.UINT16, gain=100, access_type=AccessType.READ_WRITE)
        AutomaticPowerGridReconnectionTime = Register(
            42092, "automatic_power_grid_reconnection_time", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        ComponentNameplateShortCircuitCurrent = Register(
            42093, "component_nameplate_short_circuit_current",
            DataType.UINT16, gain=100, access_type=AccessType.READ_WRITE)
        InsulationImpedanceProtectionPoint = Register(
            42097, "insulation_impedance_protection_point", DataType.UINT16,
            gain=1000, access_type=AccessType.READ_WRITE)
        VoltageUnbalanceProtectionPoint = Register(
            42098, "voltage_unbalance_protection_point", DataType.UINT16,
            gain=10, access_type=AccessType.READ_WRITE)
        PhaseProtectionPoint = Register(
            42099, "phase_protection_point", DataType.UINT16, gain=10,
            access_type=AccessType.READ_WRITE)
        PowerGridFaultStartupSoftStartTime = Register(
            42100, "power_grid_fault_startup_soft_start_time", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        CosphiPPnTriggerVoltage = Register(
            42101, "cosphi_ppn_trigger_voltage", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        CosphiPPnExitVoltage = Register(
            42102, "cosphi_ppn_exit_voltage", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        SoftStartTime = Register(
            42103, "soft_start_time", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        PowerGridFaultRecoveryAndGridConnectedTime = Register(
            42104, "power_grid_fault_recovery_and_grid_connected_time",
            DataType.UINT16, access_type=AccessType.READ_WRITE)
        ShortTimePowerGridInterruptionJudgmentTime = Register(
            42105, "short_time_power_grid_interruption_judgment_time",
            DataType.UINT32, gain=1000, count=2,
            access_type=AccessType.READ_WRITE)
        ShutdownGradient = Register(
            42107, "shutdown_gradient", DataType.UINT32, gain=1000, count=2,
            access_type=AccessType.READ_WRITE)
        LineLossCompensation = Register(
            42109, "line_loss_compensation", DataType.UINT16, gain=10,
            access_type=AccessType.READ_WRITE)
        GridFaultZeroCurrentMode = Register(
            42110, "grid_fault_zero_current_mode", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        PowerGridVoltageJumpTriggerThreshold = Register(
            42111, "power_grid_voltage_jump_trigger_threshold",
            DataType.UINT16, gain=10, access_type=AccessType.READ_WRITE)
        HVRT = Register(
            42112, "hvrt", DataType.UINT16, access_type=AccessType.READ_WRITE)
        HVRTTriggerThreshold = Register(
            42113, "hvrt_trigger_threshold", DataType.UINT16, gain=10,
            access_type=AccessType.READ_WRITE)
        HVRTPositiveReactiveCompensationFactor = Register(
            42114, "hvrt_positive_reactive_compensation_factor",
            DataType.UINT16, gain=10, access_type=AccessType.READ_WRITE)
        ShortTimePowerGridInterruptionAndQuickStartup = Register(
            42116, "short_time_power_grid_interruption_and_quick_startup",
            DataType.UINT16, access_type=AccessType.READ_WRITE)
        LVRTActiveCurrentMaintenanceCoefficient = Register(
            42118, "lvrt_active_current_maintenance_coefficient",
            DataType.UINT16, gain=100, access_type=AccessType.READ_WRITE)
        LVRT = Register(
            42119, "lvrt", DataType.UINT16, access_type=AccessType.READ_WRITE)
        LVRTTriggerThreshold = Register(
            42120, "lvrt_trigger_threshold", DataType.UINT16, gain=10,
            access_type=AccessType.READ_WRITE)
        PowerGridVoltageProtectionShieldingDuringVRT = Register(
            42121, "power_grid_voltage_protection_shielding_during_vrt",
            DataType.UINT16, access_type=AccessType.READ_WRITE)
        LVRTPositiveSequenceReactiveCompensationFactor = Register(
            42122, "lvrt_positive_sequence_reactive_compensation_factor",
            DataType.UINT16, gain=10, access_type=AccessType.READ_WRITE)
        VRTExitHysteresisThreshold = Register(
            42123, "vrt_exit_hysteresis_threshold", DataType.UINT16, gain=10,
            access_type=AccessType.READ_WRITE)
        VRTActiveCurrentLimitPercentage = Register(
            42124, "vrt_active_current_limit_percentage", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        VRTActivePowerRecoveryGradient = Register(
            42125, "vrt_active_power_recovery_gradient", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        HVRTNegativeSequenceReactivePowerCompensationFactor = Register(
            42126, "hvrt_negative_sequence_reactive_power_compensation_factor",
            DataType.UINT16, gain=10, access_type=AccessType.READ_WRITE)
        LVRTNegativeSequenceReactivePowerCompensationFactor = Register(
            42127, "lvrt_negative_sequence_reactive_power_compensation_factor",
            DataType.UINT16, gain=10, access_type=AccessType.READ_WRITE)
        PhaseAngleOffsetProtection = Register(
            42128, "phase_angle_offset_protection", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        ActiveIslandProtection = Register(
            42129, "active_island_protection", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        PassiveIslandProtection = Register(
            42130, "passive_island_protection", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        OVGRAssociatedShutdown = Register(
            42131, "ovgr_associated_shutdown", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        DryContactFunction = Register(
            42132, "dry_contact_function", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        LVRTReactiveCurrentLimitPercentage = Register(
            42133, "lvrt_reactive_current_limit_percentage", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        LVRTZeroCurrentModeThreshold = Register(
            42134, "lvrt_zero_current_mode_threshold", DataType.UINT16,
            gain=10, access_type=AccessType.READ_WRITE)
        LVRTMode = Register(
            42135, "lvrt_mode", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        VoltageRiseSuppression = Register(
            42138, "voltage_rise_suppression", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        VoltageRiseSuppressionReactivePowerAdjustmentPoint = Register(
            42139, "voltage_rise_suppression_reactive_power_adjustment_point",
            DataType.UINT16, gain=10, access_type=AccessType.READ_WRITE)
        VoltageRiseSuppressionActiveDeratingPoint = Register(
            42140, "voltage_rise_suppression_active_derating_point",
            DataType.UINT16, gain=10, access_type=AccessType.READ_WRITE)
        FMControl = Register(
            42141, "fm_control", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        FrequencyModulationControlDifferentialRate = Register(
            42142, "frequency_modulation_control_differential_rate",
            DataType.UINT16, access_type=AccessType.READ_WRITE)
        OverfrequencyDerating = Register(
            42143, "overfrequency_derating", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        OverfrequencyDeratingCutoffFrequency = Register(
            42144, "overfrequency_derating_cutoff_frequency", DataType.UINT16,
            gain=100, access_type=AccessType.READ_WRITE)
        OverfrequencyDeratingCutoffPower = Register(
            42145, "overfrequency_derating_cutoff_power", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        OverfrequencyDeratingTriggerFrequency = Register(
            42146, "overfrequency_derating_trigger_frequency",
            DataType.UINT16, gain=100, access_type=AccessType.READ_WRITE)
        OverfrequencyDeratingExitFrequency = Register(
            42147, "overfrequency_derating_exit_frequency", DataType.UINT16,
            gain=100, access_type=AccessType.READ_WRITE)
        OverfrequencyDeratingPowerRecoveryGradient = Register(
            42148, "overfrequency_derating_power_recovery_gradient",
            DataType.UINT16, access_type=AccessType.READ_WRITE)
        UnderfrequencyPowerIncrease = Register(
            42151, "underfrequency_power_increase", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        UnderfrequencyPowerRecoveryGradient = Register(
            42152, "underfrequency_power_recovery_gradient", DataType.UINT16,
            access_type=AccessType.READ_WRITE)

        # battery registers
        BatterySOC = Register(
            37760, "battery_soc", DataType.UINT16, gain=10,
            access_type=AccessType.READ)
        BatteryVoltage = Register(
            37763, "battery_voltage", DataType.UINT16, gain=10,
            access_type=AccessType.READ)
        BatteryActivePower = Register(
            37765, "battery_active_power", DataType.INT32, gain=1000, count=2,
            access_type=AccessType.READ)
        ForcedChargingAndDischargingPeriod = Register(
            47083, "forced_charging_and_discharging_period", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        ChargeFromGridFunction = Register(
            47087, "charge_from_grid_function", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        ForcibleChargeDischarge = Register(
            47100, "forcible_charge_discharge", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        TargetSOC = Register(
            47101, "target_soc", DataType.UINT16, gain=10,
            access_type=AccessType.READ_WRITE)
        ForcibleChargeDischargeSettingMode = Register(
            47246, "forcible_charge_discharge_setting_mode", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        ForcibleChargePower = Register(
            47247, "forcible_charge_power", DataType.UINT32, gain=1000,
            count=2, access_type=AccessType.READ_WRITE)
        ForcibleDischargePower = Register(
            47249, "forcible_discharge_power", DataType.UINT32, gain=1000,
            count=2, access_type=AccessType.READ_WRITE)



'''
Copyright (C) 2025  ElaadNL

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

-------------------------------------------------------------------------------

This module provides a class to interact with Victron inverters via Modbus
protocol. It defines the `VictronInverter` class, which inherits from
`ModbusBattery` and includes methods to initialize power setpoint registers,
send power and current setpoints, and retrieve various metrics such as active
power, state of charge, and phase voltage. The module also defines several
enumerations for system, battery, Multiplus GX, VEbus, and grid registers,
which are used to map Modbus registers to meaningful names and data types.

NOTE:: Slave ID's can differ based on the system configuration. Figure out the
correct slave ID's for your system by looking at the Modbus register map
provided by Victron.
'''

import warnings
from enum import Enum
from ..battery import ModbusBattery
from ..modbus import Register, AccessType, DataType


class VictronInverter(ModbusBattery):
    '''Victron inverter class

    Arguments:
        ip (str): the ip address of the victron inverter
        port (int): the port of the victron inverter
        timeout (int): the timeout of the modbus client in seconds
        capacity (int): the capacity of the battery in Wh
        max_charge_power (int): the maximum charge power of the battery in W
        max_discharge_power (int): the maximum discharge power of the battery
                                   in W
        soc (int): the state of charge of the battery in %
    '''

    def __init__(
            self,
            ip: str,
            port: int = 502,
            timeout: int = 3,
            capacity: int = None,
            max_charge_power: int = None,
            max_discharge_power: int = None,
            soc: int = None
    ):
        super().__init__(ip, port, timeout, capacity, max_charge_power,
                         max_discharge_power, soc)

    def init_power_setpoint_registers(self):
        '''Initialize power setpoint registers of the victron inverter
        Ensures that the disable charge and feedback flags are set to 0
        and the ess mode is set to 3 for external control
        '''
        self.write_register(
            self.VEbusRegisters.ess_disable_charge_flag_phase.value,
            0,  # 0 to enable charging
            self.VEbusRegisters.slave_id.value
        )
        self.write_register(
            self.VEbusRegisters.ess_disable_feedback_flag_phase.value,
            0,  # 0 to enable feedback
            self.VEbusRegisters.slave_id.value
        )
        self.write_register(
            self.SystemRegisters.ess_mode.value,
            3,  # 3 for external control
            self.SystemRegisters.slave_id.value
        )

    def set_power_setpoint_phase(self, setpoint: int, phase: int):
        '''Send the power setpoint to the victron inverter

        Arguments:
            setpoint (int): the setpoint to send in W
            phase (int): the phase to send the setpoint to
        '''
        if (self._max_charge_power is None or
                self._max_discharge_power is None):
            warnings.warn("Max charge power or max discharge power not set, \
                          setpoint may exceed limits")
        elif setpoint > self._max_charge_power:
            warnings.warn(f"Power setpoint {setpoint} exceeds the maximum \
                          charge power {self._max_charge_power} \n\
                          Reverting to maximum charge power")
            setpoint = self._max_charge_power
        elif setpoint < -self._max_discharge_power:
            warnings.warn(f"Power setpoint {setpoint} exceeds the maximum \
                          discharge power {-self._max_discharge_power} \n\
                          Reverting to maximum discharge power")
            setpoint = -self._max_discharge_power

        register = None
        if phase == 1:
            register = self.VEbusRegisters.ess_power_setpoint_phase_1.value
        elif phase == 2:
            register = self.VEbusRegisters.ess_power_setpoint_phase_2.value
        elif phase == 3:
            register = self.VEbusRegisters.ess_power_setpoint_phase_3.value
        if register:
            self.write_register(
                register,
                setpoint,
                self.VEbusRegisters.slave_id.value
            )
        else:
            raise ValueError("Invalid phase")

    def set_current_setpoint_phase(self, setpoint: int, phase: int):
        '''Send the current setpoint to the victron inverter

        Arguments:
            setpoint (int): the setpoint to send in A
            phase (int): the phase to send the setpoint to
        '''
        phase_voltage = self.get_phase_voltage(phase)
        power_setpoint = int(setpoint * phase_voltage)
        self.set_power_setpoint_phase(phase, power_setpoint)

    def set_power_setpoint(self, setpoint: int):
        '''Send the power setpoint to the victron inverter on phase 1

        Arguments:
            setpoint (int): the setpoint to send in W
        '''
        if (self._max_charge_power is None or
                self._max_discharge_power is None):
            warnings.warn("Max charge power or max discharge power not set, \
                          setpoint may exceed limits")
        elif setpoint > self._max_charge_power:
            warnings.warn(f"Power setpoint {setpoint} exceeds the maximum \
                          charge power {self._max_charge_power} \n\
                          Reverting to maximum charge power")
            setpoint = self._max_charge_power
        elif setpoint < -self._max_discharge_power:
            warnings.warn(f"Power setpoint {setpoint} exceeds the maximum \
                          discharge power {-self._max_discharge_power} \n\
                          Reverting to maximum discharge power")
            setpoint = -self._max_discharge_power

        register = self.VEbusRegisters.ess_power_setpoint_phase_1.value
        self.write_register(
            register,
            setpoint,
            self.VEbusRegisters.slave_id.value
        )

    def set_current_setpoint(self, setpoint: int):
        '''Send the current setpoint to the victron inverter on phase 1

        Arguments:
            setpoint (int): the setpoint to send in A
        '''
        phase_voltage = self.get_phase_voltage(1)
        power_setpoint = int(setpoint * phase_voltage)
        self.set_power_setpoint_phase(1, power_setpoint)

    def get_active_power_total(self):
        '''Get the total active power of the victron inverter

        Returns:
            int: the active power of the victron inverter
        '''
        return self.read_holding_register(
            self.SystemRegisters.vebus_charge_power_system.value,
            self.SystemRegisters.slave_id.value
        )

    def get_active_current_total(self):
        '''Get the total active current of the victron inverter

        Returns:
            int: the active current of the victron inverter
        '''
        return self.read_holding_register(
            self.SystemRegisters.vebus_charge_current_system.value,
            self.SystemRegisters.slave_id.value
        )

    def get_battery_soc(self):
        '''Get the state of charge of the battery

        Returns:
            int: the state of charge of the battery
        '''
        return self.read_holding_register(
            self.BatteryRegisters.state_of_charge.value,
            self.BatteryRegisters.slave_id.value
        )

    def get_phase_voltage(self, phase: int):
        '''Get the voltage of a specific phase

        Arguments:
            phase (int): the phase to get the voltage of

        Returns:
            int: the voltage of the phase
        '''
        register = None
        if phase == 1:
            register = self.VEbusRegisters.input_voltage_phase_1.value
        elif phase == 2:
            register = self.VEbusRegisters.input_voltage_phase_2.value
        elif phase == 3:
            register = self.VEbusRegisters.input_voltage_phase_3.value
        if register:
            return self.read_holding_register(
                register,
                self.VEbusRegisters.slave_id.value
            )
        else:
            raise ValueError("Invalid phase")

    def get_phase_current(self, phase: int):
        '''Get the current of a specific phase

        Arguments:
            phase (int): the phase to get the current of

        Returns:
            int: the current of the phase
        '''
        register = None
        if phase == 1:
            register = self.VEbusRegisters.input_current_phase_1.value
        elif phase == 2:
            register = self.VEbusRegisters.input_current_phase_2.value
        elif phase == 3:
            register = self.VEbusRegisters.input_current_phase_3.value
        if register:
            return self.read_holding_register(
                register,
                self.VEbusRegisters.slave_id.value
            )
        else:
            raise ValueError("Invalid phase")

    class SystemRegisters(Enum):
        slave_id = 100
        serial_system = Register(
            800, "serial system", DataType.STRING, count=6)
        ccgx_relay_1_state = Register(
            806, "ccgx relay 1 state", DataType.UINT16)
        ccgx_rSlay_2_state = Register(
            807, "ccgx relay 2 state", DataType.UINT16)
        pv_ac_coupled_on_output_l1 = Register(
            808, "pv ac coupled on output l1", DataType.UINT16)
        pv_ac_coupled_on_output_l2 = Register(
            809, "pv ac coupled on output l2", DataType.UINT16)
        pv_ac_coupled_on_output_l3 = Register(
            810, "pv ac coupled on output l3", DataType.UINT16)
        pv_ac_coupled_on_input_l1 = Register(
            811, "pv ac coupled on input l1", DataType.UINT16)
        pv_ac_coupled_on_input_l2 = Register(
            812, "pv ac coupled on input l2", DataType.UINT16)
        pv_ac_coupled_on_input_l3 = Register(
            813, "pv ac coupled on input l3", DataType.UINT16)
        pv_ac_coupled_on_generator_l1 = Register(
            814, "pv ac coupled on generator l1", DataType.UINT16)
        pv_ac_coupled_on_generator_l2 = Register(
            815, "pv ac coupled on generator l2", DataType.UINT16)
        pv_ac_coupled_on_generator_l3 = Register(
            816, "pv ac coupled on generator l3", DataType.UINT16)
        ac_consumption_l1 = Register(817, "ac consumption l1", DataType.UINT16)
        ac_consumption_l2 = Register(818, "ac consumption l2", DataType.UINT16)
        ac_consumption_l3 = Register(819, "ac consumption l3", DataType.UINT16)
        grid_l1 = Register(820, "grid l1", DataType.INT16)
        grid_l2 = Register(821, "grid l2", DataType.INT16)
        grid_l3 = Register(822, "grid l3", DataType.INT16)
        genset_l1 = Register(823, "genset l1", DataType.INT16)
        genset_l2 = Register(824, "genset l2", DataType.INT16)
        genset_l3 = Register(825, "genset l3", DataType.INT16)
        active_input_source = Register(
            826, "active input source", DataType.UINT16)
        battery_voltage_system = Register(
            840, "battery voltage system", DataType.UINT16, gain=10)
        battery_current_system = Register(
            841, "battery current system", DataType.INT16, gain=10)
        battery_power_system = Register(
            842, "battery power system", DataType.INT16)
        battery_state_of_charge_system = Register(
            843, "battery state of charge system", DataType.UINT16)
        battery_state_system = Register(
            844, "battery state system", DataType.UINT16)
        battery_consumed_amphours_system = Register(
            845, "battery consumed amphours system", DataType.UINT16)
        battery_time_to_go_system = Register(
            846, "battery time to go system", DataType.UINT16)
        pv_dc_coupled_power = Register(
            850, "pv dc coupled power", DataType.UINT16)
        pv_dc_coupled_current = Register(
            851, "pv dc coupled current", DataType.INT16)
        charger_power = Register(855, "charger power", DataType.UINT16)
        dc_system_power = Register(860, "dc system power", DataType.INT16)
        vebus_charge_current_system = Register(
            865, "vebus charge current system", DataType.INT16, gain=10)
        vebus_charge_power_system = Register(
            866, "vebus charge power system", DataType.INT16)
        inverter_charger_current = Register(
            868, "inverter charger current", DataType.INT32, gain=10)
        inverter_charger_power = Register(
            870, "inverter charger power", DataType.INT32)
        ess_mode = Register(
            2902, "ess mode", DataType.UINT16,
            access_type=AccessType.READ_WRITE)

    class BatteryRegisters(Enum):
        slave_id = 225
        battery_power = Register(258, "battery power", DataType.INT16)
        battery_voltage = Register(
            259, "battery voltage", DataType.UINT16, gain=100)
        starter_battery_voltage = Register(
            260, "starter battery voltage", DataType.UINT16, gain=100)
        current = Register(261, "current", DataType.INT16, gain=10)
        battery_temperature = Register(
            262, "battery temperature", DataType.INT16, gain=10)
        mid_point_voltage_of_the_battery_bank = Register(
            263, "mid point voltage of the battery bank", DataType.UINT16,
            gain=100)
        mid_point_deviation_of_the_battery_bank = Register(
            264, "mid point deviation of the battery bank", DataType.UINT16,
            gain=100)
        consumed_amphours = Register(
            265, "consumed amphours", DataType.UINT16, gain=-0.1)
        state_of_charge = Register(
            266, "state of charge", DataType.UINT16, gain=10)
        alarm = Register(267, "alarm", DataType.UINT16)
        low_voltage_alarm = Register(268, "low voltage alarm", DataType.UINT16)
        high_voltage_alarm = Register(
            269, "high voltage alarm", DataType.UINT16)
        low_starter_voltage_alarm = Register(
            270, "low starter voltage alarm", DataType.UINT16)
        high_starter_voltage_alarm = Register(
            271, "high starter voltage alarm", DataType.UINT16)
        low_state_of_charge_alarm = Register(
            272, "low state of charge alarm", DataType.UINT16)
        low_temperature_alarm = Register(
            273, "low temperature alarm", DataType.UINT16)
        high_temperature_alarm = Register(
            274, "high temperature alarm", DataType.UINT16)
        mid_voltage_alarm = Register(275, "mid voltage alarm", DataType.UINT16)
        low_fused_voltage_alarm = Register(
            276, "low fused voltage alarm", DataType.UINT16)
        high_fused_voltage_alarm = Register(
            277, "high fused voltage alarm", DataType.UINT16)
        fuse_blown_alarm = Register(278, "fuse blown alarm", DataType.UINT16)
        high_internal_temperature_alarm = Register(
            279, "high internal temperature alarm", DataType.UINT16)
        relay_status = Register(280, "relay status", DataType.UINT16)
        deepest_discharge = Register(
            281, "deepest discharge", DataType.UINT16, gain=-0.1)
        last_discharge = Register(
            282, "last discharge", DataType.UINT16, gain=-0.1)
        average_discharge = Register(
            283, "average discharge", DataType.UINT16, gain=-0.1)
        charge_cycles = Register(284, "charge cycles", DataType.UINT16)
        full_discharges = Register(285, "full discharges", DataType.UINT16)
        total_ah_drawn = Register(
            286, "total ah drawn", DataType.UINT16, gain=-0.1)
        minimum_voltage = Register(
            287, "minimum voltage", DataType.UINT16, gain=100)
        maximum_voltage = Register(
            288, "maximum voltage", DataType.UINT16, gain=100)
        time_since_last_full_charge = Register(
            289, "time since last full charge", DataType.UINT16, gain=0.01)
        automatic_syncs = Register(290, "automatic syncs", DataType.UINT16)
        low_voltage_alarms = Register(
            291, "low voltage alarms", DataType.UINT16)
        high_voltage_alarms = Register(
            292, "high voltage alarms", DataType.UINT16)
        low_starter_voltage_alarms = Register(
            293, "low starter voltage alarms", DataType.UINT16)
        high_starter_voltage_alarms = Register(
            294, "high starter voltage alarms", DataType.UINT16)
        minimum_starter_voltage = Register(
            295, "minimum starter voltage", DataType.UINT16, gain=100)
        maximum_starter_voltage = Register(
            296, "maximum starter voltage", DataType.UINT16, gain=100)
        low_fused_voltage_alarms = Register(
            297, "low fused voltage alarms", DataType.UINT16)
        high_fused_voltage_alarms = Register(
            298, "high fused voltage alarms", DataType.UINT16)
        minimum_fused_voltage = Register(
            299, "minimum fused voltage", DataType.UINT16, gain=100)
        maximum_fused_voltage = Register(
            300, "maximum fused voltage", DataType.UINT16, gain=100)
        discharged_energy = Register(
            301, "discharged energy", DataType.UINT16, gain=10)
        charged_energy = Register(
            302, "charged energy", DataType.UINT16, gain=10)
        time_to_go = Register(303, "time to go", DataType.UINT16, gain=0.01)
        state_of_health = Register(
            304, "state of health", DataType.UINT16, gain=10)
        max_charge_voltage = Register(
            305, "max charge voltage", DataType.UINT16, gain=10)
        min_discharge_voltage = Register(
            306, "min discharge voltage", DataType.UINT16, gain=10)
        max_charge_current = Register(
            307, "max charge current", DataType.UINT16, gain=10)
        max_discharge_current = Register(
            308, "max discharge current", DataType.UINT16, gain=10)
        capacity = Register(309, "capacity", DataType.UINT16, gain=10)
        diagnostics_1st_last_error_timestamp = Register(
            310, "diagnostics 1st last error timestamp", DataType.UINT32)
        diagnostics_2nd_last_error_timestamp = Register(
            312, "diagnostics 2nd last error timestamp", DataType.UINT32)
        diagnostics_3rd_last_error_timestamp = Register(
            314, "diagnostics 3rd last error timestamp", DataType.UINT32)
        diagnostics_4th_last_error_timestamp = Register(
            316, "diagnostics 4th last error timestamp", DataType.UINT32)
        minimum_cell_temperature = Register(
            318, "minimum cell temperature", DataType.INT16, gain=10)
        maximum_cell_temperature = Register(
            319, "maximum cell temperature", DataType.INT16, gain=10)
        high_charge_current_alarm = Register(
            320, "high charge current alarm", DataType.UINT16)
        high_discharge_current_alarm = Register(
            321, "high discharge current alarm", DataType.UINT16)
        cell_imbalance_alarm = Register(
            322, "cell imbalance alarm", DataType.UINT16)
        internal_failure_alarm = Register(
            323, "internal failure alarm", DataType.UINT16)
        high_charge_temperature_alarm = Register(
            324, "high charge temperature alarm", DataType.UINT16)
        low_charge_temperature_alarm = Register(
            325, "low charge temperature alarm", DataType.UINT16)
        low_cell_voltage_alarm = Register(
            326, "low cell voltage alarm", DataType.UINT16)
        mode = Register(327, "mode", DataType.UINT16)
        state = Register(1282, "state", DataType.UINT16)
        error = Register(1283, "error", DataType.UINT16)
        system_switch = Register(1284, "system switch", DataType.UINT16)
        balancing = Register(1285, "balancing", DataType.UINT16)
        system_number_of_batteries = Register(
            1286, "system number of batteries", DataType.UINT16)
        system_batteries_parallel = Register(
            1287, "system batteries parallel", DataType.UINT16)
        system_batteries_series = Register(
            1288, "system batteries series", DataType.UINT16)
        system_number_of_cells_per_battery = Register(
            1289, "system number of cells per battery", DataType.UINT16)
        system_minimum_cell_voltage = Register(
            1290, "system minimum cell voltage", DataType.UINT16, gain=100)
        system_maximum_cell_voltage = Register(
            1291, "system maximum cell voltage", DataType.UINT16, gain=100)
        diagnostics_shutdowns_due_to_error = Register(
            1292, "diagnostics shutdowns due to error", DataType.UINT16)
        diagnostics_1st_last_error = Register(
            1293, "diagnostics 1st last error", DataType.UINT16)
        diagnostics_2nd_last_error = Register(
            1294, "diagnostics 2nd last error", DataType.UINT16)
        diagnostics_3rd_last_error = Register(
            1295, "diagnostics 3rd last error", DataType.UINT16)
        diagnostics_4th_last_error = Register(
            1296, "diagnostics 4th last error", DataType.UINT16)
        io_allow_to_charge = Register(
            1297, "io allow to charge", DataType.UINT16)
        io_allow_to_discharge = Register(
            1298, "io allow to discharge", DataType.UINT16)
        io_external_relay = Register(
            1299, "io external relay", DataType.UINT16)
        history_min_cell_voltage = Register(
            1300, "history min cell voltage", DataType.UINT16, gain=100)
        history_max_cell_voltage = Register(
            1301, "history max cell voltage", DataType.UINT16, gain=100)

    class MultiplusGXRegisters(Enum):
        slave_id = 230
        input_voltage_phase_1 = Register(
            4500, "input voltage phase 1", DataType.UINT16)
        input_voltage_phase_2 = Register(
            4501, "input voltage phase 2", DataType.UINT16)
        input_voltage_phase_3 = Register(
            4502, "input voltage phase 3", DataType.UINT16)
        input_current_phase_1 = Register(
            4503, "input current phase 1", DataType.UINT16)
        input_current_phase_2 = Register(
            4504, "input current phase 2", DataType.UINT16)
        input_current_phase_3 = Register(
            4505, "input current phase 3", DataType.UINT16)
        input_power_phase_1 = Register(
            4506, "input power phase 1", DataType.INT16)
        input_power_phase_2 = Register(
            4507, "input power phase 2", DataType.INT16)
        input_power_phase_3 = Register(
            4508, "input power phase 3", DataType.INT16)
        input_frequency = Register(4509, "input frequency", DataType.UINT16)
        output_voltage_phase_1 = Register(
            4510, "output voltage phase 1", DataType.UINT16)
        output_voltage_phase_2 = Register(
            4511, "output voltage phase 2", DataType.UINT16)
        output_voltage_phase_3 = Register(
            4512, "output voltage phase 3", DataType.UINT16)
        output_current_phase_1 = Register(
            4513, "output current phase 1", DataType.UINT16)
        output_current_phase_2 = Register(
            4514, "output current phase 2", DataType.UINT16)
        output_current_phase_3 = Register(
            4515, "output current phase 3", DataType.UINT16)
        output_power_phase_1 = Register(
            4516, "output power phase 1", DataType.INT16)
        output_power_phase_2 = Register(
            4517, "output power phase 2", DataType.INT16)
        output_power_phase_3 = Register(
            4518, "output power phase 3", DataType.INT16)
        output_frequency = Register(4519, "output frequency", DataType.UINT16)
        ac_input_1_source_type = Register(
            4520, "ac input 1 source type", DataType.UINT16)
        ac_input_2_source_type = Register(
            4521, "ac input 2 source type", DataType.UINT16)
        ac_input_1_current_limit = Register(
            4522, "ac input 1 current limit", DataType.UINT16)
        ac_input_2_current_limit = Register(
            4523, "ac input 2 current limit", DataType.UINT16)
        phase_count = Register(4524, "phase count", DataType.UINT16)
        active_ac_input = Register(4525, "active ac input", DataType.UINT16)
        battery_voltage = Register(4526, "battery voltage", DataType.UINT16)
        battery_current = Register(4527, "battery current", DataType.INT16)
        battery_temperature = Register(
            4528, "battery temperature", DataType.INT16)
        battery_state_of_charge = Register(
            4529, "battery state of charge", DataType.UINT16)
        inverter_state = Register(4530, "inverter state", DataType.UINT16)
        switch_position = Register(4531, "switch position", DataType.UINT16)
        temperature_alarm = Register(
            4532, "temperature alarm", DataType.UINT16)
        high_voltage_alarm = Register(
            4533, "high voltage alarm", DataType.UINT16)
        high_ac_out_voltage_alarm = Register(
            4534, "high ac out voltage alarm", DataType.UINT16)
        low_battery_temperature_alarm = Register(
            4535, "low battery temperature alarm", DataType.UINT16)
        low_voltage_alarm = Register(
            4536, "low voltage alarm", DataType.UINT16)
        low_ac_out_voltage_alarm = Register(
            4537, "low ac out voltage alarm", DataType.UINT16)
        overload_alarm = Register(4538, "overload alarm", DataType.UINT16)
        high_dc_ripple_alarm = Register(
            4539, "high dc ripple alarm", DataType.UINT16)
        pv_power = Register(4540, "pv power", DataType.UINT16)
        user_yield = Register(4541, "user yield", DataType.UINT16)
        relay_on_the_multi_rs = Register(
            4542, "relay on the multi rs", DataType.UINT16)
        mpp_operation_mode = Register(
            4543, "mpp operation mode", DataType.UINT16)
        pv_voltage = Register(4544, "pv voltage", DataType.UINT16)
        error_code = Register(4545, "error code", DataType.UINT16)
        energy_from_ac_in_1_to_ac_out = Register(
            4546, "energy from ac in 1 to ac out", DataType.UINT32)
        energy_from_ac_in_1_to_battery = Register(
            4548, "energy from ac in 1 to battery", DataType.UINT32)
        energy_from_ac_in_2_to_ac_out = Register(
            4550, "energy from ac in 2 to ac out", DataType.UINT32)
        energy_from_ac_in_2_to_battery = Register(
            4552, "energy from ac in 2 to battery", DataType.UINT32)
        energy_from_ac_out_to_ac_in_1 = Register(
            4554, "energy from ac out to ac in 1", DataType.UINT32)
        energy_from_ac_out_to_ac_in_2 = Register(
            4556, "energy from ac out to ac in 2", DataType.UINT32)
        energy_from_battery_to_ac_in_1 = Register(
            4558, "energy from battery to ac in 1", DataType.UINT32)
        energy_from_battery_to_ac_in_2 = Register(
            4560, "energy from battery to ac in 2", DataType.UINT32)
        energy_from_battery_to_ac_out = Register(
            4562, "energy from battery to ac out", DataType.UINT32)
        energy_from_ac_out_to_battery = Register(
            4564, "energy from ac out to battery", DataType.UINT32)
        energy_from_solar_to_ac_in_1 = Register(
            4566, "energy from solar to ac in 1", DataType.UINT32)
        energy_from_solar_to_ac_in_2 = Register(
            4568, "energy from solar to ac in 2", DataType.UINT32)
        energy_from_solar_to_ac_out = Register(
            4570, "energy from solar to ac out", DataType.UINT32)
        energy_from_solar_to_battery = Register(
            4572, "energy from solar to battery", DataType.UINT32)
        yield_today = Register(
            4574, "yield today", DataType.UINT16)
        maximum_charge_power_today = Register(
            4575, "maximum charge power today", DataType.UINT16)
        yield_yesterday = Register(
            4576, "yield yesterday", DataType.UINT16)
        maximum_charge_power_yesterday = Register(
            4577, "maximum charge power yesterday", DataType.UINT16)
        yield_today_for_tracker_0 = Register(
            4578, "yield today for tracker 0", DataType.UINT16)
        yield_today_for_tracker_1 = Register(
            4579, "yield today for tracker 1", DataType.UINT16)
        yield_today_for_tracker_2 = Register(
            4580, "yield today for tracker 2", DataType.UINT16)
        yield_today_for_tracker_3 = Register(
            4581, "yield today for tracker 3", DataType.UINT16)
        yield_yesterday_for_tracker_0 = Register(
            4582, "yield yesterday for tracker 0", DataType.UINT16)
        yield_yesterday_for_tracker_1 = Register(
            4583, "yield yesterday for tracker 1", DataType.UINT16)
        yield_yesterday_for_tracker_2 = Register(
            4584, "yield yesterday for tracker 2", DataType.UINT16)
        yield_yesterday_for_tracker_3 = Register(
            4585, "yield yesterday for tracker 3", DataType.UINT16)
        maximum_charge_power_today_for_tracker_0 = Register(
            4586, "maximum charge power today for tracker 0", DataType.UINT16)
        maximum_charge_power_today_for_tracker_1 = Register(
            4587, "maximum charge power today for tracker 1", DataType.UINT16)
        maximum_charge_power_today_for_tracker_2 = Register(
            4588, "maximum charge power today for tracker 2", DataType.UINT16)
        maximum_charge_power_today_for_tracker_3 = Register(
            4589, "maximum charge power today for tracker 3", DataType.UINT16)
        maximum_charge_power_yesterday_tracker_0 = Register(
            4590, "maximum charge power yesterday tracker 0", DataType.UINT16)
        maximum_charge_power_yesterday_tracker_1 = Register(
            4591, "maximum charge power yesterday tracker 1", DataType.UINT16)
        maximum_charge_power_yesterday_tracker_2 = Register(
            4592, "maximum charge power yesterday tracker 2", DataType.UINT16)
        maximum_charge_power_yesterday_tracker_3 = Register(
            4593, "maximum charge power yesterday tracker 3", DataType.UINT16)
        pv_voltage_for_tracker_0 = Register(
            4594, "pv voltage for tracker 0", DataType.UINT16)
        pv_voltage_for_tracker_1 = Register(
            4595, "pv voltage for tracker 1", DataType.UINT16)
        pv_voltage_for_tracker_2 = Register(
            4596, "pv voltage for tracker 2", DataType.UINT16)
        pv_voltage_for_tracker_3 = Register(
            4597, "pv voltage for tracker 3", DataType.UINT16)
        pv_power_for_tracker_0 = Register(
            4598, "pv power for tracker 0", DataType.UINT16)
        pv_power_for_tracker_1 = Register(
            4599, "pv power for tracker 1", DataType.UINT16)
        pv_power_for_tracker_2 = Register(
            4600, "pv power for tracker 2", DataType.UINT16)
        pv_power_for_tracker_3 = Register(
            4601, "pv power for tracker 3", DataType.UINT16)

    class VEbusRegisters(Enum):
        slave_id = 227
        input_voltage_phase_1 = Register(
            3, "input voltage phase 1", DataType.UINT16, gain=10)
        input_voltage_phase_2 = Register(
            4, "input voltage phase 2", DataType.UINT16, gain=10)
        input_voltage_phase_3 = Register(
            5, "input voltage phase 3", DataType.UINT16, gain=10)
        input_current_phase_1 = Register(
            6, "input current phase 1", DataType.INT16, gain=10)
        input_current_phase_2 = Register(
            7, "input current phase 2", DataType.INT16, gain=10)
        input_current_phase_3 = Register(
            8, "input current phase 3", DataType.INT16, gain=10)
        input_frequency_1 = Register(
            9, "input frequency 1", DataType.INT16, gain=100)
        input_frequency_2 = Register(
            10, "input frequency 2", DataType.INT16, gain=100)
        input_frequency_3 = Register(
            11, "input frequency 3", DataType.INT16, gain=100)
        input_power_phase_1 = Register(
            12, "input power phase 1", DataType.INT16, gain=0.1)
        input_power_phase_2 = Register(
            13, "input power phase 2", DataType.INT16, gain=0.1)
        input_power_phase_3 = Register(
            14, "input power phase 3", DataType.INT16, gain=0.1)
        output_voltage_phase_1 = Register(
            15, "output voltage phase 1", DataType.UINT16)
        output_voltage_phase_2 = Register(
            16, "output voltage phase 2", DataType.UINT16)
        output_voltage_phase_3 = Register(
            17, "output voltage phase 3", DataType.UINT16)
        output_current_phase_1 = Register(
            18, "output current phase 1", DataType.INT16)
        output_current_phase_2 = Register(
            19, "output current phase 2", DataType.INT16)
        output_current_phase_3 = Register(
            20, "output current phase 3", DataType.INT16)
        output_frequency = Register(21, "output frequency", DataType.UINT16)
        active_input_current_limit = Register(
            22, "active input current limit", DataType.INT16)
        output_power_phase_1 = Register(
            23, "output power phase 1", DataType.INT16, gain=0.1)
        output_power_phase_2 = Register(
            24, "output power phase 2", DataType.INT16, gain=0.1)
        output_power_phase_3 = Register(
            25, "output power phase 3", DataType.INT16, gain=0.1)
        battery_voltage = Register(26, "battery voltage", DataType.UINT16)
        battery_current = Register(27, "battery current", DataType.INT16)
        phase_count = Register(28, "phase count", DataType.UINT16)
        active_input = Register(29, "active input", DataType.UINT16)
        vebus_state_of_charge = Register(
            30, "vebus state of charge", DataType.UINT16)
        vebus_state = Register(31, "vebus state", DataType.UINT16)
        vebus_error = Register(32, "vebus error", DataType.UINT16)
        switch_position = Register(33, "switch position", DataType.UINT16)
        temperature_alarm = Register(34, "temperature alarm", DataType.UINT16)
        low_battery_alarm = Register(35, "low battery alarm", DataType.UINT16)
        overload_alarm = Register(36, "overload alarm", DataType.UINT16)
        ess_power_setpoint_phase_1 = Register(
            37, "ess power setpoint phase 1", DataType.INT16,
            access_type=AccessType.READ_WRITE)
        ess_disable_charge_flag_phase = Register(
            38, "ess disable charge flag phase", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        ess_disable_feedback_flag_phase = Register(
            39, "ess disable feedback flag phase", DataType.UINT16,
            access_type=AccessType.READ_WRITE)
        ess_power_setpoint_phase_2 = Register(
            40, "ess power setpoint phase 2", DataType.INT16,
            access_type=AccessType.READ_WRITE)
        ess_power_setpoint_phase_3 = Register(
            41, "ess power setpoint phase 3", DataType.INT16,
            access_type=AccessType.READ_WRITE)
        temperature_sensor_alarm = Register(
            42, "temperature sensor alarm", DataType.UINT16)
        voltage_sensor_alarm = Register(
            43, "voltage sensor alarm", DataType.UINT16)
        temperature_alarm_l1 = Register(
            44, "temperature alarm l1", DataType.UINT16)
        low_battery_alarm_l1 = Register(
            45, "low battery alarm l1", DataType.UINT16)
        overload_alarm_l1 = Register(46, "overload alarm l1", DataType.UINT16)
        ripple_alarm_l1 = Register(47, "ripple alarm l1", DataType.UINT16)
        temperature_alarm_l2 = Register(
            48, "temperature alarm l2", DataType.UINT16)
        low_battery_alarm_l2 = Register(
            49, "low battery alarm l2", DataType.UINT16)
        overload_alarm_l2 = Register(50, "overload alarm l2", DataType.UINT16)
        ripple_alarm_l2 = Register(51, "ripple alarm l2", DataType.UINT16)
        temperature_alarm_l3 = Register(
            52, "temperature alarm l3", DataType.UINT16)
        low_battery_alarm_l3 = Register(
            53, "low battery alarm l3", DataType.UINT16)
        overload_alarm_l3 = Register(54, "overload alarm l3", DataType.UINT16)
        ripple_alarm_l3 = Register(55, "ripple alarm l3", DataType.UINT16)
        disable_pv_inverter = Register(
            56, "disable pv inverter", DataType.UINT16)
        vebus_bms_allows_battery_to_be_charged = Register(
            57, "vebus bms allows battery to be charged", DataType.UINT16)
        vebus_bms_allows_battery_to_be_discharged = Register(
            58, "vebus bms allows battery to be discharged", DataType.UINT16)
        vebus_bms_is_expected = Register(
            59, "vebus bms is expected", DataType.UINT16)
        vebus_bms_error = Register(60, "vebus bms error", DataType.UINT16)
        battery_temperature = Register(
            61, "battery temperature", DataType.INT16)
        vebus_reset = Register(62, "vebus reset", DataType.UINT16)
        phase_rotation_warning = Register(
            63, "phase rotation warning", DataType.UINT16)
        grid_lost_alarm = Register(64, "grid lost alarm", DataType.UINT16)
        feed_dc_overvoltage_into_grid = Register(
            65, "feed dc overvoltage into grid", DataType.UINT16)
        maximum_overvoltage_feed_in_power_l1 = Register(
            66, "maximum overvoltage feed in power l1", DataType.UINT16)
        maximum_overvoltage_feed_in_power_l2 = Register(
            67, "maximum overvoltage feed in power l2", DataType.UINT16)
        maximum_overvoltage_feed_in_power_l3 = Register(
            68, "maximum overvoltage feed in power l3", DataType.UINT16)

    class GridRegisters(Enum):
        grid_l1_power = Register(2600, "grid l1 power", DataType.INT16)
        grid_l2_power = Register(2601, "grid l2 power", DataType.INT16)
        grid_l3_power = Register(2602, "grid l3 power", DataType.INT16)
        grid_l1_energy_from_net = Register(
            2603, "grid l1 energy from net", DataType.UINT16, gain=100)
        grid_l2_energy_from_net = Register(
            2604, "grid l2 energy from net", DataType.UINT16, gain=100)
        grid_l3_energy_from_net = Register(
            2605, "grid l3 energy from net", DataType.UINT16, gain=100)
        grid_l1_energy_to_net = Register(
            2606, "grid l1 energy to net", DataType.UINT16, gain=100)
        grid_l2_energy_to_net = Register(
            2607, "grid l2 energy to net", DataType.UINT16, gain=100)
        grid_l3_energy_to_net = Register(
            2608, "grid l3 energy to net", DataType.UINT16, gain=100)
        serial = Register(2609, "serial", DataType.STRING, count=14)
        grid_l1_voltage = Register(
            2616, "grid l1 voltage", DataType.UINT16, gain=10)
        grid_l1_current = Register(
            2617, "grid l1 current", DataType.INT16, gain=10)
        grid_l2_voltage = Register(
            2618, "grid l2 voltage", DataType.UINT16, gain=10)
        grid_l2_current = Register(
            2619, "grid l2 current", DataType.INT16, gain=10)
        grid_l3_voltage = Register(
            2620, "grid l3 voltage", DataType.UINT16, gain=10)
        grid_l3_current = Register(
            2621, "grid l3 current", DataType.INT16, gain=10)
        grid_l1_energy_from_net_2 = Register(
            2622, "grid l1 energy from net", DataType.UINT32, gain=100)
        grid_l2_energy_from_net_2 = Register(
            2624, "grid l2 energy from net", DataType.UINT32, gain=100)
        grid_l3_energy_from_net_2 = Register(
            2626, "grid l3 energy from net", DataType.UINT32, gain=100)
        grid_l1_energy_to_net_2 = Register(
            2628, "grid l1 energy to net", DataType.UINT32, gain=100)
        grid_l2_energy_to_net_2 = Register(
            2630, "grid l2 energy to net", DataType.UINT32, gain=100)
        grid_l3_energy_to_net_2 = Register(
            2632, "grid l3 energy to net", DataType.UINT32, gain=100)



"""
Copyright (C) 2025  ElaadNL

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

-------------------------------------------------------------------------------

This module provides a set of classes and functions to interact with Modbus
devices. It includes definitions for Modbus registers, data types, and access
types, as well as a ModbusDevice class that facilitates reading from and
writing to Modbus devices using the pymodbus library.

Classes:
    AccessType: Enum representing the access types of a Modbus register.
    DataType: Enum representing the data types of a Modbus register.
    Register: Dataclass representing a Modbus register.
    ModbusDevice: Class representing a Modbus device and providing methods to
                  interact with it.

Usage:
    Create an instance of the ModbusDevice class with the IP address and port
    of the Modbus device. Use the provided methods to read from and write to
    the Modbus device registers.
"""


from enum import Enum
from dataclasses import dataclass

from pymodbus.client import ModbusTcpClient
from pymodbus.payload import BinaryPayloadDecoder
from pymodbus.payload import BinaryPayloadBuilder
from pymodbus.constants import Endian


class AccessType(Enum):
    READ = "r"
    WRITE = "w"
    READ_WRITE = "rw"


class DataType(Enum):
    INT16 = "int16"
    UINT16 = "uint16"
    INT32 = "int32"
    UINT32 = "uint32"
    INT64 = "int64"
    UINT64 = "uint64"
    STRING = "string"


@dataclass
class Register:
    '''Represents a modbus register

    Arguments:
        address (int): the address of the register
        name (str): the name of the register
        data_type (DataType): the data type of the register
        byte_order (Endian): the byte order of the register
        count (int): the number of registers
        gain (float): the gain of the register
        unit (str): the unit of the register
        access_type (AccessType): the access type of the register
        mapping (dict): the mapping of the register
    '''
    address: int
    name: str = None
    data_type: DataType = DataType.INT16
    byte_order: Endian = Endian.BIG
    count: int = 1
    gain: float = 1
    unit: str = None
    access_type: AccessType = AccessType.READ
    mapping: dict = None

    def __post_init__(self):
        if self.data_type in [DataType.INT32, DataType.UINT32]:
            self.count = 2
        elif self.data_type in [DataType.INT64, DataType.UINT64]:
            self.count = 4

    def __str__(self):
        return f"Register: {self.address}, {self.name}, {self.data_type}, \
                {self.byte_order}, {self.count}, {self.gain}, {self.unit}, \
                {self.access_type}, {self.mapping}"

    def decode(self, value: any):
        '''Decodes the value of the register

        Arguments:
            value: the value of the register

        Returns:
            the decoded value of the register
        '''
        if self.data_type is None:
            raise ValueError(f"Data type not set, please provide register \
                             with a data type! \n {self}")

        decoder = BinaryPayloadDecoder.fromRegisters(value, self.byte_order)
        if self.data_type == DataType.INT16:
            return decoder.decode_16bit_int() / self.gain
        elif self.data_type == DataType.UINT16:
            return decoder.decode_16bit_uint() / self.gain
        elif self.data_type == DataType.INT32:
            return decoder.decode_32bit_int() / self.gain
        elif self.data_type == DataType.UINT32:
            return decoder.decode_32bit_uint() / self.gain
        elif self.data_type == DataType.INT64:
            return decoder.decode_64bit_int() / self.gain
        elif self.data_type == DataType.UINT64:
            return decoder.decode_64bit_uint() / self.gain
        elif self.data_type == DataType.STRING:
            return decoder.decode_string(self.count * 2)

    def build_payload(self, value: any):
        '''Builds the payload of the register

        Arguments:
            value: the value of the register

        Returns:
            the payload of the register
        '''
        if self.data_type is None:
            raise ValueError(f"Data type not set, please provide register \
                             with a data type! \n {self}")

        builder = BinaryPayloadBuilder(byteorder=self.byte_order)
        if self.data_type == DataType.STRING:
            builder.add_string(value)
        else:
            value = int(value * self.gain)
            if self.data_type == DataType.INT16:
                builder.add_16bit_int(value)
            elif self.data_type == DataType.UINT16:
                builder.add_16bit_uint(value)
            elif self.data_type == DataType.INT32:
                builder.add_32bit_int(value)
            elif self.data_type == DataType.UINT32:
                builder.add_32bit_uint(value)
            elif self.data_type == DataType.INT64:
                builder.add_64bit_int(value)
            elif self.data_type == DataType.UINT64:
                builder.add_64bit_uint(value)
        return builder.to_registers()


class ModbusDevice:
    '''Represents a modbus device

    Arguments:
        modbus_ip (str): ip address of the modbus device
        modbus_port (int): port of the modbus device
        timeout (int): connection request timeout of the modbus device in
                       seconds
    '''

    def __init__(
            self,
            modbus_ip: str,
            modbus_port: int = 502,
            timeout: int = 3
    ):
        self._modbus_ip = modbus_ip
        self._modbus_port = modbus_port
        self._timeout = timeout
        self._client = ModbusTcpClient(
            self._modbus_ip,
            port=self._modbus_port,
            timeout=timeout
        )
        self.connect()

    def connect(self):
        '''Connects to the modbus device'''
        if not self._client.connected:
            self._client.connect()
            assert self._client.connected

    def disconnect(self):
        '''Disconnects from the modbus device'''
        self._client.close()

    def read_coils(self, address: int, count: int = 1, slave_id: int = 1):
        '''Reads coils (code 0x01) from the modbus battery

        Arguments:
            address (int): the address of the coils
            count (int): the number of coils to read
            slave_id (int): the id of the slave
        '''
        if not self._client.connected:
            raise Exception("Client not connected")

        response = self._client.read_coils(address, count, slave_id)
        return response.bits

    def read_discrete_inputs(self, address: int, count: int = 1,
                             slave_id: int = 1):
        '''Reads discrete inputs (code 0x02) from the modbus battery

        Arguments:
            address (int): the address of the discrete inputs
            count (int): the number of discrete inputs to read
            slave_id (int): the id of the slave
        '''
        if not self._client.connected:
            raise Exception("Client not connected")

        response = self._client.read_discrete_inputs(
            address, count, slave_id)
        return response.bits

    def read_holding_register(self, register: Register, slave_id: int = 1):
        '''Reads holding registers (code 0x03) from the modbus battery

        Arguments:
            register (Register): the register to read
            slave_id (int): the id of the slave

        Returns:
            the decoded value of the register
        '''
        if not self._client.connected:
            raise Exception("Client not connected")
        if (register.access_type != AccessType.READ and
                register.access_type != AccessType.READ_WRITE):
            raise Exception(f"Register {register.name} does not have read \
                            access")

        response = self._client.read_holding_registers(
            register.address, register.count, slave_id)
        if response.isError():
            print("Error reading holding register")
            print(response)
            return None
        return register.decode(response.registers)

    def read_holding_register_raw(self, address: int, count: int = 1,
                                  slave_id: int = 1):
        '''Reads holding registers (code 0x03) from the modbus battery

        Arguments:
            address (int): the address of the holding registers
            count (int): the number of holding registers to read
            slave_id (int): the id of the slave

        Returns:
            the raw modbus response
        '''
        if not self._client.connected:
            raise Exception("Client not connected")

        response = self._client.read_holding_registers(
            address, count, slave_id)
        if response.isError():
            print("Error reading holding register")
            print(response)
            return None
        return response.registers

    def read_input_register(self, register: Register, slave_id: int = 1):
        '''Reads input registers (code 0x04) from the modbus battery

        Arguments:
            register (Register): the register to read
            slave_id (int): the id of the slave
        '''
        if not self._client.connected:
            raise Exception("Client not connected")
        if (register.access_type != AccessType.READ and
                register.access_type != AccessType.READ_WRITE):
            raise Exception(f"Register {register.name} does not have read \
                            access")

        response = self._client.read_holding_registers(
            register.address, register.count, slave_id)
        if response.isError():
            print("Error reading holding register")
            print(response)
            return None
        return register.decode(response.registers)

    def read_input_register_raw(self, address: int, count: int = 1,
                                slave_id: int = 1):
        '''Reads input registers (code 0x04) from the modbus battery

        Arguments:
            address (int): the address of the input registers
            count (int): the number of input registers to read
            slave_id (int): the id of the slave
        '''
        if not self._client.connected:
            raise Exception("Client not connected")

        response = self._client.read_input_registers(
            address, count, slave_id)
        if response.isError():
            print("Error reading holding register")
            print(response)
            return None
        return response.registers

    def write_coil(self, address: int, value: bool, slave_id: int = 1):
        '''Writes a single coil (code 0x05) to the modbus battery

        Arguments:
            address (int): the address of the coil
            value (bool): the value of the coil
            slave_id (int): the id of the slave
        '''
        if not self._client.connected:
            raise Exception("Client not connected")

        self._client.write_coil(address, value, slave_id)

    def write_register(self, register: Register, value: any,
                       slave_id: int = 1, force_0x10: bool = False):
        '''Writes a register (code 0x06 / 0x10) to the modbus battery using
        BinaryPayloadBuilder to build the payload

        Arguments:
            register (Register): the register to write
            value (int): the value of the register
            slave_id (int): the id of the slave
            force_0x10 (bool): force the use of function code 0x10
        '''
        if not self._client.connected:
            raise Exception("Client not connected")
        if (register.access_type != AccessType.WRITE and
                register.access_type != AccessType.READ_WRITE):
            raise Exception(f"Register {register.name} does not have write \
                            access")

        if register.count > 1:
            payload = register.build_payload(value)
            self._client.write_registers(
                register.address, payload, slave_id)
        else:
            payload = register.build_payload(value)
            if force_0x10:
                self._client.write_registers(
                    register.address, payload[0], slave_id)
            else:
                self._client.write_register(
                    register.address, payload[0], slave_id)

    def write_register_raw(self, address: int, value: int, slave_id: int = 1):
        '''Writes a single register (code 0x06) to the modbus battery

        Arguments:
            address (int): the address of the register
            value (int): the value of the register
            slave_id (int): the id of the slave
        '''
        if not self._client.connected:
            raise Exception("Client not connected")

        self._client.write_register(address, value, slave_id)

    def write_coils(self, address: int, values: list[bool], slave_id: int = 1):
        '''Writes multiple coils (code 0x0F) to the modbus battery

        Arguments:
            address (int): the address of the coils
            values (list): the values of the coils
            slave_id (int): the id of the slave
        '''
        if not self._client.connected:
            raise Exception("Client not connected")

        self._client.write_coils(address, values, slave_id)

    def write_registers(self, registers: list[Register], values: list[any],
                        slave_id: int = 1, force_0x10: bool = False):
        '''Writes multiple registers (code 0x10) to the modbus battery using
        BinaryPayloadBuilder to build the payload

        Arguments:
            registers (list): the registers to write
            values (list): the values of the registers
            slave_id (int): the id of the slave
            force_0x10 (bool): force the use of function code 0x10
        '''
        if not self._client.connected:
            raise Exception("Client not connected")

        for register, value in zip(registers, values):
            if (register.access_type != AccessType.WRITE and
                    register.access_type != AccessType.READ_WRITE):
                raise Exception(f"Register {register.name} does not have \
                                write access")

            if register.count > 1:
                payload = register.build_payload(value)
                self._client.write_registers(
                    register.address, payload, slave_id)
            else:
                payload = register.build_payload(value)
                if force_0x10:
                    self._client.write_registers(
                        register.address, payload[0], slave_id)
                else:
                    self._client.write_register(
                        register.address, payload[0], slave_id)

    def write_registers_raw(self, address: int, values: list[int],
                            slave_id: int = 1):
        '''Writes multiple registers (code 0x10) to the modbus battery

        Arguments:
            address (int): start address of the registers
            values (list): the values of the registers
            slave_id (int): the id of the slave
        '''
        if not self._client.connected:
            raise Exception("Client not connected")

        self._client.write_registers(address, values, slave_id)

    def read_all_registers(self, registers: list[Register], slave_id: int = 1):
        '''Reads all registers and returns a dict of all register names and
        values

        Arguments:
            registers (list): the registers to read
            slave_id (int): the id of the slave

        Returns:
            dict of all register names and values
        '''
        if not self._client.connected:
            raise Exception("Client not connected")

        values = {}

        for register in registers:
            if isinstance(register, Register):
                if (register.access_type == AccessType.READ or
                        register.access_type == AccessType.READ_WRITE):
                    values[register.name] = self.read_holding_register(
                        register, slave_id)
                else:
                    values[register.name] = "No read access"

        return values



'''
Copyright (C) 2025  ElaadNL

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

-------------------------------------------------------------------------------

This module provides a class to interface with a Deye inverter using Modbus
protocol. The DeyeInverter class allows for communication with the inverter
to read and write various parameters, such as power setpoints, battery state
of charge, and grid voltage.

The class utilizes the ModbusBattery base class and defines specific registers
for the Deye inverter. It includes methods to set power and current setpoints,
read all registers, and get specific values like active power and battery SOC.
'''

import warnings
from enum import Enum
from ..battery import ModbusBattery
from ..modbus import Register, AccessType

# Inverter constants
CURRENT_LIMIT = 195


class DeyeModbusInverter(ModbusBattery):
    '''Deye modbus inverter class
    this class is used to communicate with a deye inverter through the
    solarman data logger(LSE-3)

    Arguments:
        data_logger_ip (str): the ip address of the solarman data logger
        data_logger_port (int): the port of the solarman data logger
        timeout (int): the timeout of the modbus client in seconds
        manual_mode (bool): whether or not to use manual mode to set the power
                            setpoint
        capacity (int): the capacity of the battery
        max_charge_power (int): the maximum charge power of the battery in W
        max_discharge_power (int): the maximum discharge power of the battery
                                   in W
        soc (int): the state of charge of the battery in %

    Note:
        The setpoint is calculated using the internal voltage of the battery
        and does not automatically update when the voltage changes. in order to
        update the setpoint, you can simply call the setpoint method again
        with the same setpoint.
    '''

    def __init__(
            self,
            data_logger_ip: str,
            data_logger_port: int = 8899,
            timeout: int = 3,
            manual_mode: bool = True,
            capacity: int = None,
            max_charge_power: int = None,
            max_discharge_power: int = None,
            soc: int = None
    ):
        super().__init__(data_logger_ip, data_logger_port, timeout, capacity,
                         max_charge_power, max_discharge_power, soc)

        self._manual_mode = manual_mode
        if manual_mode:
            self._init_power_setpoint_registers()

    def _set_discharge_setpoint(self, setpoint: int):
        # Set discharge power setpoint
        self.write_register(
            self.HoldingRegisters.SellModeKWPoint6.value, setpoint)
        # Set SOC setpoint to 5% to discharge
        self.write_register(
            self.HoldingRegisters.SellModeBattCapacity6.value, 5)

    def _set_charge_setpoint(self, setpoint: int):
        voltage = self.read_holding_register(
            self.HoldingRegisters.BatteryVoltage.value, 1)

        max_charge_current = int(setpoint / voltage)
        if max_charge_current > CURRENT_LIMIT:
            max_charge_current = CURRENT_LIMIT

        self.write_register(
            self.HoldingRegisters.MaxChargeCurrent.value, max_charge_current)
        # Set SOC setpoint to 100% to charge
        self.write_register(
            self.HoldingRegisters.SellModeBattCapacity6.value, 100)

    def _update_charge_current(self, setpoint: int):
        voltage = self.read_holding_register(
            self.HoldingRegisters.BatteryVoltage.value, 1)

        max_charge_current = int(setpoint / voltage)
        print(max_charge_current)

        if max_charge_current > CURRENT_LIMIT:
            max_charge_current = CURRENT_LIMIT

        self.write_register(
            self.HoldingRegisters.MaxChargeCurrent.value, max_charge_current)

    def _init_power_setpoint_registers(self):
        # set TimeOfUse to 0x00FF
        # this enables the time of use settings for all days of the week
        self.write_register(
            self.HoldingRegisters.TimeOfUse.value, 0x00FF)

        self.write_register(
            self.HoldingRegisters.SellTimePoint1.value, 100)
        self.write_register(
            self.HoldingRegisters.SellTimePoint2.value, 100)
        self.write_register(
            self.HoldingRegisters.SellTimePoint3.value, 100)
        self.write_register(
            self.HoldingRegisters.SellTimePoint4.value, 100)
        self.write_register(
            self.HoldingRegisters.SellTimePoint5.value, 100)
        self.write_register(
            self.HoldingRegisters.SellTimePoint6.value, 100)

        # with all time frames set to the same time, the inverter will always
        # use the settings from the sixth time frame so we only need to update
        # the settings in the sixth time frame to change the power setpoint

        self.write_register(
            self.HoldingRegisters.ChargeModeEnable6.value, 1)

    def set_power_setpoint(self, setpoint: int):
        '''Sets the power setpoint of the inverter

        Arguments:
            setpoint (int): the power setpoint to set in W
        '''
        if (self._max_charge_power is None or
                self._max_discharge_power is None):
            warnings.warn("Max charge power or max discharge power not set, \
                          setpoint may exceed limits")
        elif setpoint > self._max_charge_power:
            warnings.warn(f"Power setpoint {setpoint} exceeds the maximum \
                          charge power {self._max_charge_power} \n\
                          Reverting to maximum charge power")
            setpoint = self._max_charge_power
        elif setpoint < -self._max_discharge_power:
            warnings.warn(f"Power setpoint {setpoint} exceeds the maximum \
                          discharge power {-self._max_discharge_power} \n\
                          Reverting to maximum discharge power")
            setpoint = -self._max_discharge_power

        if setpoint > 0:
            self._set_charge_setpoint(setpoint)
        else:
            self._set_discharge_setpoint(abs(setpoint))

    def set_current_setpoint(self, setpoint: int):
        '''Sets the current setpoint of the inverter

        Arguments:
            setpoint (int): the current setpoint to set in A
        '''
        phase_voltage = self.get_phase_voltage(1)
        power_setpoint = setpoint * phase_voltage
        self.set_power_setpoint(power_setpoint)

    def get_active_power_total(self):
        '''Gets the active power of the inverter

        Returns:
            float: the active power of the inverter
        '''
        return self.read_holding_register(
            self.HoldingRegisters.ActivePower.value, 1)

    def get_active_current_total(self):
        '''Gets the active total current of the inverter

        Returns:
            float: the active current of the inverter
        '''
        power = self.get_active_power_total()
        voltage = self.read_holding_register(
            self.HoldingRegisters.GridphaseVoltageA.value, 1)

        return power / voltage

    def get_battery_soc(self):
        '''Gets the state of charge of the battery

        Returns:
            float: the state of charge of the battery
        '''
        return self.read_holding_register(
            self.HoldingRegisters.BatterySOC.value, 1)

    def get_phase_voltage(self, phase: int):
        '''Gets the grid voltage
        Arguments:
            phase (int): the phase to get the voltage from
        Returns:
            float: the grid voltage
        '''
        register = None
        if phase == 1:
            register = self.HoldingRegisters.GridphaseVoltageA.value
        elif phase == 2:
            register = self.HoldingRegisters.GridphaseVoltageB.value
        elif phase == 3:
            register = self.HoldingRegisters.GridphaseVoltageC.value
        if register:
            return self.read_holding_register(register, 1)
        else:
            raise ValueError("Invalid phase")

    class HoldingRegisters(Enum):
        LimitControlFunction = Register(
            142, "LimitControlFunction", access_type=AccessType.READ_WRITE)
        LimitMaxGridPower = Register(
            143, "LimitMaxGridPower", unit="W",
            access_type=AccessType.READ_WRITE)
        TimeOfUse = Register(
            146, "TimeOfUse", access_type=AccessType.READ_WRITE)

        # Starting time for each time frame
        SellTimePoint1 = Register(
            148, "SellTimePoint1", unit="hh:mm",
            access_type=AccessType.READ_WRITE)
        SellTimePoint2 = Register(
            149, "SellTimePoint2", unit="hh:mm",
            access_type=AccessType.READ_WRITE)
        SellTimePoint3 = Register(
            150, "SellTimePoint3", unit="hh:mm",
            access_type=AccessType.READ_WRITE)
        SellTimePoint4 = Register(
            151, "SellTimePoint4", unit="hh:mm",
            access_type=AccessType.READ_WRITE)
        SellTimePoint5 = Register(
            152, "SellTimePoint5", unit="hh:mm",
            access_type=AccessType.READ_WRITE)
        SellTimePoint6 = Register(
            153, "SellTimePoint6", unit="hh:mm",
            access_type=AccessType.READ_WRITE)

        # Discharge power setpoint
        SellModeKWPoint1 = Register(
            154, "SellModeKWPoint1", unit="W",
            access_type=AccessType.READ_WRITE)
        SellModeKWPoint2 = Register(
            155, "SellModeKWPoint2", unit="W",
            access_type=AccessType.READ_WRITE)
        SellModeKWPoint3 = Register(
            156, "SellModeKWPoint3", unit="W",
            access_type=AccessType.READ_WRITE)
        SellModeKWPoint4 = Register(
            157, "SellModeKWPoint4", unit="W",
            access_type=AccessType.READ_WRITE)
        SellModeKWPoint5 = Register(
            158, "SellModeKWPoint5", unit="W",
            access_type=AccessType.READ_WRITE)
        SellModeKWPoint6 = Register(
            159, "SellModeKWPoint6", unit="W",
            access_type=AccessType.READ_WRITE)

        # Battery voltage setpoint can be used instead of SOC setpoint
        SellModeBattVolt1 = Register(
            160, "SellModeBattVolt1", unit="V",
            access_type=AccessType.READ_WRITE)
        SellModeBattVolt2 = Register(
            161, "SellModeBattVolt2", unit="V",
            access_type=AccessType.READ_WRITE)
        SellModeBattVolt3 = Register(
            162, "SellModeBattVolt3", unit="V",
            access_type=AccessType.READ_WRITE)
        SellModeBattVolt4 = Register(
            163, "SellModeBattVolt4", unit="V",
            access_type=AccessType.READ_WRITE)
        SellModeBattVolt5 = Register(
            164, "SellModeBattVolt5", unit="V",
            access_type=AccessType.READ_WRITE)
        SellModeBattVolt6 = Register(
            165, "SellModeBattVolt6", unit="V",
            access_type=AccessType.READ_WRITE)

        # SOC setpoint
        SellModeBattCapacity1 = Register(
            166, "SellModeBattCapacity1", unit="%",
            access_type=AccessType.READ_WRITE)
        SellModeBattCapacity2 = Register(
            167, "SellModeBattCapacity2", unit="%",
            access_type=AccessType.READ_WRITE)
        SellModeBattCapacity3 = Register(
            168, "SellModeBattCapacity3", unit="%",
            access_type=AccessType.READ_WRITE)
        SellModeBattCapacity4 = Register(
            169, "SellModeBattCapacity4", unit="%",
            access_type=AccessType.READ_WRITE)
        SellModeBattCapacity5 = Register(
            170, "SellModeBattCapacity5", unit="%",
            access_type=AccessType.READ_WRITE)
        SellModeBattCapacity6 = Register(
            171, "SellModeBattCapacity6", unit="%",
            access_type=AccessType.READ_WRITE)

        # Enable/disable whether or not to allow charging the battery from the
        # grid
        ChargeModeEnable1 = Register(
            172, "ChargeModeEnable1", access_type=AccessType.READ_WRITE)
        ChargeModeEnable2 = Register(
            173, "ChargeModeEnable2", access_type=AccessType.READ_WRITE)
        ChargeModeEnable3 = Register(
            174, "ChargeModeEnable3", access_type=AccessType.READ_WRITE)
        ChargeModeEnable4 = Register(
            175, "ChargeModeEnable4", access_type=AccessType.READ_WRITE)
        ChargeModeEnable5 = Register(
            176, "ChargeModeEnable5", access_type=AccessType.READ_WRITE)
        ChargeModeEnable6 = Register(
            177, "ChargeModeEnable6", access_type=AccessType.READ_WRITE)

        MaxChargeCurrent = Register(
            108, "MaxChargeCurrent", unit="A",
            access_type=AccessType.READ_WRITE)
        MaxDischargeCurrent = Register(
            109, "MaxDischargeCurrent", unit="A",
            access_type=AccessType.READ_WRITE)

        BatterySOC = Register(
            214, "BatterySOC", unit="%", access_type=AccessType.READ)
        BatteryVoltage = Register(
            215, "BatteryVoltage", gain=100, unit="V",
            access_type=AccessType.READ)

        BatteryOutputCurrent = Register(
            591, "BatteryOutputCurrent", gain=100, unit="A",
            access_type=AccessType.READ)
        GridphaseVoltageA = Register(
            598, "GridphaseVoltageA", gain=10, unit="V",
            access_type=AccessType.READ)
        GridphaseVoltageB = Register(
            599, "GridphaseVoltageB", gain=10, unit="V",
            access_type=AccessType.READ)
        GridphaseVoltageC = Register(
            600, "GridphaseVoltageC", gain=10, unit="V",
            access_type=AccessType.READ)

        ActivePower = Register(
            607, "ActivePower", unit="W", access_type=AccessType.READ)

        GridSideInnerCurrentA = Register(
            610, "GridSideInnerCurrentA", gain=100, unit="A",
            access_type=AccessType.READ)
        GridSideInnerCurrentB = Register(
            611, "GridSideInnerCurrentB", gain=100, unit="A",
            access_type=AccessType.READ)
        GridSideInnerCurrentC = Register(
            612, "GridSideInnerCurrentC", gain=100, unit="A",
            access_type=AccessType.READ)
        OutOfGridCurrentA = Register(
            613, "OutOfGridCurrentA", gain=100, unit="A",
            access_type=AccessType.READ)
        OutOfGridCurrentB = Register(
            614, "OutOfGridCurrentB", gain=100, unit="A",
            access_type=AccessType.READ)
        OutOfGridCurrentC = Register(
            615, "OutOfGridCurrentC", gain=100, unit="A",
            access_type=AccessType.READ)

        InverterOutputCurrentA = Register(
            630, "InverterOutputCurrentA", gain=100, unit="A",
            access_type=AccessType.READ)
        InverterOutputCurrentB = Register(
            631, "InverterOutputCurrentB", gain=100, unit="A",
            access_type=AccessType.READ)
        InverterOutputCurrentC = Register(
            632, "InverterOutputCurrentC", gain=100, unit="A",
            access_type=AccessType.READ)



'''
SMA Inverter module
Copyright (C) 2025  ElaadNL

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

-------------------------------------------------------------------------------

This module defines the SmaInverter class, which represents a battery system
with various functionalities such as retrieving power status, state of charge
(SOC), and grid voltage. It also allows setting the power setpoint and current
setpoint of the inverter. The SmaInverter class inherits from the ModbusBattery
class and uses the Modbus protocol to interact with the inverter.

'''

from ..battery import ModbusBattery
from ..modbus import Register, DataType, AccessType
from enum import Enum


class SmaInverter(ModbusBattery):
    '''Sma inverter class

    Arguments:
        ip (str): the ip address of the victron inverter
        port (int): the port of the victron inverter
        timeout (int): the timeout of the modbus client in seconds
        capacity (int): the capacity of the battery in Wh
        max_charge_power (int): the maximum charge power of the battery in W
        max_discharge_power (int): the maximum discharge power of the battery
                                   in W
        soc (int): the state of charge of the battery in %
    '''

    def __init__(
            self,
            ip: str,
            port: int = 502,
            timeout: int = 3,
            capacity: int = None,
            max_charge_power: int = None,
            max_discharge_power: int = None,
            soc: int = None
    ):
        super().__init__(ip, port, timeout, capacity, max_charge_power,
                         max_discharge_power, soc)
        # Set the power control to 0x0322 to enable active power control
        self.write_register(self.SmaRegisters.power_control.value, 0x0322, 3)

    def set_power_setpoint(self, setpoint: int):
        '''Set the power setpoint of the inverter

        Arguments:
            setpoint (int): the power setpoint in W
        '''
        self.write_register(
            self.SmaRegisters.active_power_setpoint.value, setpoint, 3)

    def set_current_setpoint(self, setpoint: int):
        '''Set the current setpoint of the inverter

        Arguments:
            setpoint (int): the current setpoint in A
        '''
        phase_voltage = self.get_phase_voltage(1)
        power_setpoint = int(setpoint * phase_voltage)
        self.set_power_setpoint(power_setpoint)

    def get_active_power_total(self):
        '''Get the total active power of the inverter
        '''
        return self.read_holding_register(
            self.SmaRegisters.active_power.value, 3)

    def get_active_current_total(self):
        '''Get the total active current of the inverter
        '''
        return self.get_active_power_total() / self.get_phase_voltage(1)

    def get_battery_soc(self):
        '''Get the state of charge of the battery
        '''
        return self.read_holding_register(
            self.SmaRegisters.battery_soc.value, 3)

    def get_phase_voltage(self, phase: int):
        '''Get the grid voltage of the inverter

        Arguments:
            phase (int): the phase number
        '''
        if phase == 1:
            return self.read_holding_register(
                self.SmaRegisters.grid_voltage_l1.value, 3)
        elif phase == 2:
            return self.read_holding_register(
                self.SmaRegisters.grid_voltage_l2.value, 3)
        elif phase == 3:
            return self.read_holding_register(
                self.SmaRegisters.grid_voltage_l3.value, 3)
        else:
            raise ValueError("Invalid phase number")

    def get_grid_current(self, phase: int):
        '''Get the grid current of the inverter

        Arguments:
            phase (int): the phase number
        '''
        if phase == 1:
            return self.read_holding_register(
                self.SmaRegisters.grid_current_l1.value, 3)
        elif phase == 2:
            return self.read_holding_register(
                self.SmaRegisters.grid_current_l2.value, 3)
        elif phase == 3:
            return self.read_holding_register(
                self.SmaRegisters.grid_current_l3.value, 3)
        else:
            raise ValueError("Invalid phase number")

    class SmaRegisters(Enum):
        active_power_setpoint = Register(
            40149, "active_power_setpoint", DataType.INT32,
            count=2, access_type=AccessType.WRITE)
        power_control = Register(
            40151, "power_control", DataType.UINT32,
            count=2, access_type=AccessType.WRITE)
        operating_mode_power_setpoint = Register(
            40210, "operating_mode_power_setpoint", DataType.UINT32,
            count=2, access_type=AccessType.READ_WRITE)
        battery_soc = Register(
            30845, "battery_soc", DataType.UINT32, count=2)
        active_power = Register(
            30775, "active_power", DataType.INT32, count=2)
        battery_current = Register(
            30843, "battery_current", DataType.INT32, count=2, gain=1000)
        battery_temperature = Register(
            30849, "battery_temperature", DataType.UINT32, count=2, gain=10)
        battery_voltage = Register(
            30851, "battery_voltage", DataType.UINT32, count=2, gain=100)
        operating_status = Register(
            33003, "operating_status", DataType.UINT32, count=2)
        grid_voltage_l1 = Register(
            30783, "grid_voltage_l1", DataType.UINT32, count=2, gain=100)
        grid_voltage_l2 = Register(
            30785, "grid_voltage_l2", DataType.UINT32, count=2, gain=100)
        grid_voltage_l3 = Register(
            30787, "grid_voltage_l3", DataType.UINT32, count=2, gain=100)
        grid_current_l1 = Register(
            30977, "grid_current_l1", DataType.INT32, count=2, gain=1000)
        grid_current_l2 = Register(
            30979, "grid_current_l2", DataType.INT32, count=2, gain=1000)
        grid_current_l3 = Register(
            30981, "grid_current_l3", DataType.INT32, count=2, gain=1000)
